##############################################################################bl
# MIT License
#
# Copyright (c) 2021 - 2025 Advanced Micro Devices, Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
##############################################################################el

import ctypes
import glob
import json
import math
import os
import re
import shutil
import sys
import threading
from abc import abstractmethod
from pathlib import Path

import pandas as pd
import yaml

import wafer_core.lib.rocprofiler.compute.gui.config as config
from wafer_core.lib.rocprofiler.compute.gui.utils.logger import (
    console_debug,
    console_error,
    console_log,
    console_warning,
    demarcate,
)
from wafer_core.lib.rocprofiler.compute.gui.utils.mi_gpu_spec import mi_gpu_specs
from wafer_core.lib.rocprofiler.compute.gui.utils.parser import build_in_vars, supported_denom
from wafer_core.lib.rocprofiler.compute.gui.utils.utils import (
    add_counter_extra_config_input_yaml,
    capture_subprocess_output,
    convert_metric_id_to_panel_idx,
    detect_rocprof,
    get_submodules,
    is_tcc_channel_counter,
    using_v3,
)


class OmniSoC_Base:
    def __init__(
        self, args, mspec
    ):  # new info field will contain rocminfo or sysinfo to populate properties
        console_debug("[omnisoc init]")
        self.__args = args
        self.__arch = None
        self._mspec = mspec
        self.__perfmon_dir = None
        self.__perfmon_config = (
            {}
        )  # Per IP block max number of simulutaneous counters. GFX IP Blocks
        self.__soc_params = {}  # SoC specifications
        self.__compatible_profilers = []  # Store profilers compatible with SoC
        self.populate_mspec()

    def __hash__(self):
        return hash(self.__arch)

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.__arch == other.get_soc()

    def set_perfmon_dir(self, path: str):
        self.__perfmon_dir = path

    def set_perfmon_config(self, config: dict):
        self.__perfmon_config = config

    def get_soc_param(self):
        return self.__soc_params

    def set_arch(self, arch: str):
        self.__arch = arch

    def get_arch(self):
        return self.__arch

    def get_args(self):
        return self.__args

    def set_compatible_profilers(self, profiler_names: list):
        self.__compatible_profilers = profiler_names

    def get_compatible_profilers(self):
        return self.__compatible_profilers

    def populate_mspec(self):
        from wafer_core.lib.rocprofiler.compute.gui.utils.specs import run, search, total_sqc

        if not hasattr(self._mspec, "_rocminfo") or self._mspec._rocminfo is None:
            return

        # load stats from rocminfo
        self._mspec.gpu_l1 = ""
        self._mspec.gpu_l2 = ""
        for idx2, linetext in enumerate(self._mspec._rocminfo):
            key = search(r"^\s*L1:\s+ ([a-zA-Z0-9]+)\s*", linetext)
            if key != None:
                self._mspec.gpu_l1 = key
                continue

            key = search(r"^\s*L2:\s+ ([a-zA-Z0-9]+)\s*", linetext)
            if key != None:
                self._mspec.gpu_l2 = key
                continue

            key = search(r"^\s*Max Clock Freq\. \(MHz\):\s+([0-9]+)", linetext)
            if key != None:
                self._mspec.max_sclk = key
                continue

            key = search(r"^\s*Compute Unit:\s+ ([a-zA-Z0-9]+)\s*", linetext)
            if key != None:
                self._mspec.cu_per_gpu = key
                continue

            key = search(r"^\s*SIMDs per CU:\s+ ([a-zA-Z0-9]+)\s*", linetext)
            if key != None:
                self._mspec.simd_per_cu = key
                continue

            key = search(r"^\s*Shader Engines:\s+ ([a-zA-Z0-9]+)\s*", linetext)
            if key != None:
                self._mspec.se_per_gpu = key
                continue

            key = search(r"^\s*Wavefront Size:\s+ ([a-zA-Z0-9]+)\s*", linetext)
            if key != None:
                self._mspec.wave_size = key
                continue

            key = search(r"^\s*Workgroup Max Size:\s+ ([a-zA-Z0-9]+)\s*", linetext)
            if key != None:
                self._mspec.workgroup_max_size = key
                continue

            key = search(r"^\s*Max Waves Per CU:\s+ ([a-zA-Z0-9]+)\s*", linetext)
            if key != None:
                self._mspec.max_waves_per_cu = key
                break

        self._mspec.sqc_per_gpu = str(
            total_sqc(
                self._mspec.gpu_arch, self._mspec.cu_per_gpu, self._mspec.se_per_gpu
            )
        )

        # Parse json from amd-smi static --clock
        amd_smi_mclk = run(["amd-smi", "static", "--clock", "--json"], exit_on_error=True)
        amd_smi_mclk = json.loads(amd_smi_mclk)

        if isinstance(amd_smi_mclk, dict):
            # The output of `amd-smi static --clock --json` is a dict with amd-smi>=26.0.0.
            amd_smi_mclk = amd_smi_mclk["gpu_data"][0]["clock"]["mem"]["frequency_levels"]
        else:
            # For backward compatibility: the output of `amd-smi static --clock --json` used to be a list for amd-smi<26.0.0.
            amd_smi_mclk = amd_smi_mclk[0]["clock"]["mem"]["frequency_levels"]

        # Choose the highest level of memory clock frequency
        amd_smi_mclk = amd_smi_mclk[sorted(amd_smi_mclk.keys())[-1]]
        # 100 Mhz -> 100
        self._mspec.max_mclk = amd_smi_mclk.split(" ")[0]

        console_debug("max mem clock is {}".format(self._mspec.max_mclk))

        # these are just max's now, because the parsing was broken and this was inconsistent
        # with how we use the clocks elsewhere (all max, all the time)
        self._mspec.cur_sclk = self._mspec.max_sclk
        self._mspec.cur_mclk = self._mspec.max_mclk

        self._mspec.gpu_series = mi_gpu_specs.get_gpu_series(self._mspec.gpu_arch)
        # specify gpu model name for gfx942 hardware
        self._mspec.gpu_model = mi_gpu_specs.get_gpu_model(
            self._mspec.gpu_arch, self._mspec.gpu_chip_id
        )

        if not self._mspec.gpu_model:
            self._mspec.gpu_model = self.detect_gpu_model(self._mspec.gpu_arch)

        self._mspec.num_xcd = str(
            mi_gpu_specs.get_num_xcds(
                self._mspec.gpu_arch, self._mspec.gpu_model, self._mspec.compute_partition
            )
        )

    @demarcate
    def detect_gpu_model(self, gpu_arch):
        """
        Detects the GPU model using various identifiers from 'amd-smi static'.
        Falls back through multiple methods if the primary method fails.
        """

        from utils.specs import run, search

        # TODO: use amd-smi python api when available
        amd_smi_static = run(["amd-smi", "static", "--gpu=0"], exit_on_error=True)

        # Purposely search for patterns without variants suffix to try and match a known GPU model.
        detection_methods = [
            {
                "name": "Market Name",
                "pattern": r"MARKET_NAME:\s*.*(mi|MI\d*[a-zA-Z]*)",
            },
            {
                "name": "VBIOS Name",
                "pattern": r"NAME:\s*.*(mi|MI\d*[a-zA-Z]*)",
            },
            {"name": "Product Name", "pattern": r"PRODUCT_NAME:\s*.*(mi|MI\d*[a-zA-Z]*)"},
        ]

        gpu_model = None
        for method in detection_methods:
            console_log(f"Determining GPU model using {method['name']}.")
            gpu_model = search(method["pattern"], amd_smi_static)
            if gpu_model:
                break

        if not gpu_model:
            console_warning("Unable to determine the GPU model.")
            return

        gpu_model = self._adjust_mi300_model(gpu_model.lower(), gpu_arch.lower())

        if gpu_model.lower() not in mi_gpu_specs.get_num_xcds_dict().keys():
            console_warning(f"Unknown GPU model detected: '{gpu_model}'.")
            return

        return gpu_model.upper()

    def _adjust_mi300_model(self, gpu_model, gpu_arch):
        """
        Applies specific adjustments for MI300 series GPU models based on architecture.
        """

        if gpu_model in ["mi300a", "mi300x"]:
            if gpu_arch in ["gfx940", "gfx941"]:
                gpu_model += "_a0"
            elif gpu_arch == "gfx942":
                gpu_model += "_a1"

        return gpu_model

    @demarcate
    def detect_counters(self):
        """
        Create a set of counters required for the selected report sections.
        Parse analysis report configuration files based on the selected report sections to be filtered.
        """
        counters = set()
        config_filenames = {
            filename: []
            for filename in os.listdir(
                Path(self.get_args().config_dir).joinpath(self.__arch)
            )
            if filename.endswith(".yaml")
        }
        metric_ids = [
            name
            for name, type in self.get_args().filter_blocks.items()
            if type == "metric_id"
        ]
        file_ids = []
        for section in metric_ids:
            section_num = convert_metric_id_to_panel_idx(section)
            file_id = str(section_num // 100)
            # Convert "4" to "04"
            if len(file_id) == 1:
                file_id = f"0{file_id}"
            file_ids.append(file_id)
            # Apply sub section filtering
            for config_filename in config_filenames:
                if config_filename.startswith(file_id) and section_num % 100:
                    config_filenames[config_filename].append(section_num)

        # Apply section filters only if metric ids have been provided for filtering
        if metric_ids:
            # Identify yaml files corresponding to file_ids
            config_filenames = {
                filename: subsections
                for filename, subsections in config_filenames.items()
                if filename.startswith(tuple(file_ids))
            }

        for config_filename, subsections in config_filenames.items():
            # Read the yaml file
            with open(
                Path(self.get_args().config_dir).joinpath(self.__arch, config_filename),
                "r",
            ) as stream:
                section_config = yaml.safe_load(stream)
            # Extract subsection if section is of the form 4.52
            if subsections:
                section_config_text = "\n".join(
                    [
                        # Convert yaml to string
                        yaml.dump(subsection, sort_keys=False)
                        for subsection in section_config["Panel Config"]["data source"]
                        if subsection["metric_table"]["id"] in subsections
                    ]
                )
            else:
                # Convert yaml to string
                section_config_text = yaml.dump(section_config, sort_keys=False)
            counters = counters.union(self.parse_counters(section_config_text))

        # Handle TCC channel counters: if hw_counter_matches has elements ending with '['
        # Expand and interleve the TCC channel counters
        # e.g.  TCC_HIT[0] TCC_ATOMIC[0] ... TCC_HIT[1] TCC_ATOMIC[1] ...
        num_xcd_for_pmc_file = 1
        if using_v3():
            num_xcd_for_pmc_file = int(self._mspec.num_xcd)

        for counter_name in counters.copy():
            if counter_name.startswith("TCC") and counter_name.endswith("["):
                counters.remove(counter_name)
                counter_name = counter_name.split("[")[0]
                counters = counters.union(
                    {
                        f"{counter_name}[{i}]"
                        for i in range(num_xcd_for_pmc_file * int(self._mspec._l2_banks))
                    }
                )

        return counters

    @demarcate
    def perfmon_filter(self, roofline_perfmon_only: bool):
        """Filter default performance counter set based on user arguments"""
        if (
            roofline_perfmon_only
            and Path(self.get_args().path).joinpath("pmc_perf.csv").is_file()
        ):
            return

        if roofline_perfmon_only:
            counters = set()
            for fname in glob.glob(self.__perfmon_dir + "/" + "pmc_roof_perf.txt"):
                lines = open(fname, "r").read().splitlines()
                for line in lines:
                    # Strip all comments, skip empty lines
                    stext = line.split("#")[0].strip()
                    if not stext:
                        continue
                    # all pmc counters start with  "pmc:"
                    m = re.match(r"^pmc:(.*)", stext)
                    if m is None:
                        continue
                    # de-duplicate counters
                    counters = counters.union(set(m.group(1).split()))
        else:
            counters = self.detect_counters()
            # Perfmon hardware block filtering
            filter_hardware_blocks = [
                name
                for name, type in self.get_args().filter_blocks.items()
                if type == "hardware_block"
            ]
            if filter_hardware_blocks:
                counters = {
                    counter_name
                    for counter_name in counters
                    if counter_name.startswith(tuple(filter_hardware_blocks))
                }

        if not using_v3():
            # Counters not supported in rocprof v1 / v2
            counters = counters - {"SQ_INSTS_VALU_MFMA_F8", "SQ_INSTS_VALU_MFMA_MOPS_F8"}

        # Following counters are not supported
        # TCP_TCP_LATENCY_sum (except for gfx950)
        # SQC_DCACHE_INFLIGHT_LEVEL
        counters = counters - {"SQC_DCACHE_INFLIGHT_LEVEL"}
        if self.__arch != "gfx950":
            counters = counters - {"TCP_TCP_LATENCY_sum"}

        # SQ_ACCUM_PREV_HIRES will be injected for level counters later on
        counters = counters - {"SQ_ACCUM_PREV_HIRES"}

        # Coalesce and writeback workload specific perfmon
        self.perfmon_coalesce(counters)

    @demarcate
    def parse_counters(self, config_text):
        """
        Create a set of all hardware counters mentioned in the given config file content string
        """
        hw_counter_matches, variable_matches = self.parse_counters_text(config_text)

        # get hw counters and variables for all supported denominators
        for formula in supported_denom.values():
            hw_counter_matches_denom, variable_matches_denom = self.parse_counters_text(
                formula
            )
            hw_counter_matches.update(hw_counter_matches_denom)
            variable_matches.update(variable_matches_denom)

        # get hw counters corresponding to variables recursively
        while variable_matches:
            subvariable_matches = set()
            for var in variable_matches:
                if var in build_in_vars:
                    hw_counter_matches_vars, variable_matches_vars = (
                        self.parse_counters_text(build_in_vars[var])
                    )
                    hw_counter_matches.update(hw_counter_matches_vars)
                    subvariable_matches.update(variable_matches_vars)
            # process new found variables
            variable_matches = subvariable_matches - variable_matches

        return hw_counter_matches

    def parse_counters_text(self, text):
        """Parse out hardware counters and variables from given text"""
        # hw counter name should start with ip block name
        # hw counter name should have all capital letters or digits and should not end with underscore
        # he counter name can either optionally end with '[' or '_sum'
        hw_counter_regex = (
            r"(?:SQ|SQC|TA|TD|TCP|TCC|CPC|CPF|SPI|GRBM)_[0-9A-Z_]*[0-9A-Z](?:\[|_sum)*"
        )
        # only capture the variable name after $ using capturing group
        variable_regex = r"\$([0-9A-Za-z_]*[0-9A-Za-z])"
        hw_counter_matches = set(re.findall(hw_counter_regex, text))
        variable_matches = set(re.findall(variable_regex, text))
        # variable matches cannot be counters
        hw_counter_matches = hw_counter_matches - variable_matches
        return hw_counter_matches, variable_matches

    def get_rocprof_supported_counters(self):
        rocprof_cmd = detect_rocprof(self.get_args())

        if rocprof_cmd != "rocprofiler-sdk":
            console_warning(
                "rocprof v1 / v2 / v3 interfaces will be removed in favor of "
                "rocprofiler-sdk interface in a future release. To use rocprofiler-sdk "
                "interface, please set the environment variable ROCPROF to 'rocprofiler-sdk' "
                "and optionally provide the path to librocprofiler-sdk.so library via the "
                "--rocprofiler-sdk-library-path option."
            )

        rocprof_counters = set()

        if str(rocprof_cmd).endswith("rocprof"):
            command = [rocprof_cmd, "--list-basic"]
            success, output = capture_subprocess_output(command, enable_logging=False)
            # return code should be 1 so success should be False
            if success:
                console_error(
                    f"Failed to list rocprof supported counters using command: {command}"
                )
            for line in output.splitlines():
                if "gpu-agent" in line:
                    counters, _ = self.parse_counters_text(line.split(":")[1].strip())
                    rocprof_counters.update(counters)

            command = [rocprof_cmd, "--list-derived"]
            success, output = capture_subprocess_output(command, enable_logging=False)
            # return code should be 1 so success should be False
            if success:
                console_error(
                    f"Failed to list rocprof supported counters using command: {command}"
                )
            for line in output.splitlines():
                if "gpu-agent" in line:
                    counters, _ = self.parse_counters_text(line.split(":")[1].strip())
                    rocprof_counters.update(counters)

        elif str(rocprof_cmd).endswith("rocprofv2"):
            command = [rocprof_cmd, "--list-counters"]
            success, output = capture_subprocess_output(command, enable_logging=False)
            # return code should be 1 so success should be False
            if success:
                console_error(
                    f"Failed to list rocprof supported counters using command: {command}"
                )
            for line in output.splitlines():
                if "gfx" in line:
                    counters, _ = self.parse_counters_text(line.split(":")[2].strip())
                    rocprof_counters.update(counters)

        elif str(rocprof_cmd).endswith("rocprofv3"):
            command = [rocprof_cmd, "--list-avail"]
            success, output = capture_subprocess_output(command, enable_logging=False)
            # return code should be 0 so success should be True
            if not success:
                console_error(
                    f"Failed to list rocprof supported counters using command: {command}"
                )
            for line in output.splitlines():
                if "counter_name" in line:
                    counters, _ = self.parse_counters_text(line.split(":")[1].strip())
                    rocprof_counters.update(counters)
            # Custom counter support for mi100 for rocprofv3
            if self._mspec.gpu_model.lower() == "mi100":
                counter_defs_path = (
                    config.rocprof_compute_home
                    / "rocprof_compute_soc"
                    / "profile_configs"
                    / "gfx908_counter_defs.yaml"
                )
                with open(counter_defs_path, "r") as fp:
                    counter_defs_contents = fp.read()
                counters, _ = self.parse_counters_text(counter_defs_contents)
                rocprof_counters.update(counters)

        elif str(rocprof_cmd) == "rocprofiler-sdk":
            sys.path.append(
                str(
                    Path(self.get_args().rocprofiler_sdk_library_path).parent.parent
                    / "bin"
                )
            )
            from rocprofv3_avail_module import avail

            avail.loadLibrary.libname = str(
                Path(self.get_args().rocprofiler_sdk_library_path).parent.parent
                / "libexec"
                / "rocprofiler-sdk"
                / "librocprofv3-list-avail.so"
            )
            counters = avail.get_counters()
            rocprof_counters = {
                counter.name
                for counter in counters[list(counters.keys())[0]]
                if hasattr(counter, "block") or hasattr(counter, "expression")
            }
            # Custom counter support for mi100 for rocprofiler-sdk
            if self._mspec.gpu_model.lower() == "mi100":
                counter_defs_path = (
                    config.rocprof_compute_home
                    / "rocprof_compute_soc"
                    / "profile_configs"
                    / "gfx908_counter_defs.yaml"
                )
                with open(counter_defs_path, "r") as fp:
                    counter_defs_contents = fp.read()
                counters, _ = self.parse_counters_text(counter_defs_contents)
                rocprof_counters.update(counters)

        else:
            console_error(
                "Incompatible profiler: %s. Supported profilers include: %s"
                % (rocprof_cmd, get_submodules("rocprof_compute_profile"))
            )

        return rocprof_counters

    @demarcate
    def perfmon_coalesce(self, counters):
        """Sort and bucket all related performance counters to minimize required application passes"""

        # Create workload directory
        # In some cases (i.e. --specs) path will not be given
        if hasattr(self.get_args(), "path"):
            if self.get_args().path == str(Path(os.getcwd()).joinpath("workloads")):
                workload_dir = str(
                    Path(self.get_args().path).joinpath(
                        self.get_args().name, self._mspec.gpu_model
                    )
                )
            else:
                workload_dir = self.get_args().path

        # Initialize directories
        if not Path(workload_dir).is_dir():
            os.makedirs(workload_dir)
        elif not Path(workload_dir).is_symlink():
            shutil.rmtree(workload_dir)
        else:
            os.unlink(workload_dir)

        workload_perfmon_dir = workload_dir + "/perfmon"
        os.makedirs(workload_perfmon_dir)

        # Sanity check whether counters are supported by underlying rocprof tool
        rocprof_counters = self.get_rocprof_supported_counters()
        # rocprof does not support TCC channel counters, so remove channel suffix for comparison
        not_supported_counters = {
            counter.split("[")[0] if is_tcc_channel_counter(counter) else counter
            for counter in counters
        } - rocprof_counters
        if not_supported_counters:
            console_warning(
                f"Following counters might not be supported by rocprof: {', '.join(not_supported_counters)} "
            )
        # We might be providing definitions of unsupported counters, so still try to collect them
        if not counters:
            console_error(
                "profiling",
                "No performance counters to collect, please check the provided profiling filters",
            )
        else:
            console_debug(f"Collecting following counters: {', '.join(counters)} ")

        output_files = []

        accu_file_count = 0
        # Create separate perfmon file for LEVEL counters without _sum suffix
        # TCC LEVEL counters are handled channel wise, so ignore them
        for counter in counters.copy():
            if (
                "LEVEL" in counter
                and not counter.endswith("_sum")
                and not is_tcc_channel_counter(counter)
            ):
                counters.remove(counter)
                output_files.append(CounterFile(counter + ".txt", self.__perfmon_config))
                output_files[-1].add(counter)
                if using_v3():
                    # v3 does not support SQ_ACCUM_PREV_HIRES. Instead we defined our own
                    # counters in counter_defs.yaml that use the accumulate() function. These
                    # use the name of the accumulate counter with _ACCUM appended to them.
                    output_files[-1].add(counter + "_ACCUM")
                else:
                    output_files[-1].add("SQ_ACCUM_PREV_HIRES")
                accu_file_count += 1

        file_count = 0
        # Store all channels for a TCC channel counter in the same file
        tcc_channel_counter_file_map = dict()
        # Store all pipes for SPI pipe counters in the same file
        spi_pipe_counter_file_map = dict()
        for ctr in counters:
            # Store all channels for a TCC channel counter in the same file
            if is_tcc_channel_counter(ctr):
                output_file = tcc_channel_counter_file_map.get(ctr.split("[")[0])
                if output_file:
                    output_file.add(ctr)
                    continue
            # Add counter to first file that has room
            added = False
            for i in range(len(output_files)):
                if output_files[i].add(ctr):
                    added = True
                    # Store all channels for a TCC channel counter in the same file
                    if is_tcc_channel_counter(ctr):
                        tcc_channel_counter_file_map[ctr.split("[")[0]] = output_files[i]
                    break

            # All files are full, create a new file
            if not added:
                output_files.append(
                    CounterFile(
                        "pmc_perf_{}.txt".format(file_count), self.__perfmon_config
                    )
                )
                file_count += 1
                output_files[-1].add(ctr)

        console_debug("profiling", "perfmon_coalesce file_count %s" % file_count)

        # TODO: rewrite the above logic for spatial_multiplexing later
        if self.get_args().spatial_multiplexing:

            # TODO: more error checking
            if len(self.get_args().spatial_multiplexing) != 3:
                console_error(
                    "profiling",
                    "multiplexing need provide node_idx node_count and gpu_count",
                )

            node_idx = int(self.get_args().spatial_multiplexing[0])
            node_count = int(self.get_args().spatial_multiplexing[1])
            gpu_count = int(self.get_args().spatial_multiplexing[2])

            old_group_num = file_count + accu_file_count
            new_bucket_count = node_count * gpu_count
            groups_per_bucket = math.ceil(
                old_group_num / new_bucket_count
            )  # It equals to file num per node
            max_groups_per_node = groups_per_bucket * gpu_count

            group_start = node_idx * max_groups_per_node
            group_end = min((node_idx + 1) * max_groups_per_node, old_group_num)

            console_debug(
                "profiling",
                "spatial_multiplexing node_idx %s, node_count %s, gpu_count: %s, old_group_num %s, "
                "new_bucket_count %s, groups_per_bucket %s, max_groups_per_node %s, "
                "group_start %s, group_end %s"
                % (
                    node_idx,
                    node_count,
                    gpu_count,
                    old_group_num,
                    new_bucket_count,
                    groups_per_bucket,
                    max_groups_per_node,
                    group_start,
                    group_end,
                ),
            )

            for f_idx in range(groups_per_bucket):
                file_name = str(
                    Path(workload_perfmon_dir).joinpath(
                        "pmc_perf_" + "node_" + str(node_idx) + "_" + str(f_idx) + ".txt"
                    )
                )

                pmc = []
                for g_idx in range(
                    group_start + f_idx * gpu_count,
                    min(group_end, group_start + (f_idx + 1) * gpu_count),
                ):
                    gpu_idx = g_idx % gpu_count
                    for block_name in output_files[g_idx].blocks.keys():
                        for ctr in output_files[g_idx].blocks[block_name].elements:
                            pmc.append(ctr + ":device=" + str(gpu_idx))

                stext = "pmc: " + " ".join(pmc)

                # Write counters to file
                fd = open(file_name, "w")
                fd.write(stext + "\n\n")
                fd.close()

        else:
            # Output to files
            for f in output_files:
                file_name_txt = str(Path(workload_perfmon_dir).joinpath(f.file_name_txt))
                file_name_yaml = str(
                    Path(workload_perfmon_dir).joinpath(f.file_name_yaml)
                )

                pmc = []
                counter_def = dict()
                for ctr in [
                    ctr
                    for block_name in f.blocks
                    for ctr in f.blocks[block_name].elements
                ]:
                    pmc.append(ctr)
                    if using_v3():
                        # MI 100 accumulate counters dont work with rocprofiler sdk
                        if self._mspec.gpu_model.lower() != "mi100":
                            # Add accumulation counters definitions
                            if ctr == "SQ_IFETCH_LEVEL":
                                counter_def = add_counter_extra_config_input_yaml(
                                    counter_def,
                                    "SQ_IFETCH_LEVEL_ACCUM",
                                    "SQ_IFETCH_LEVEL accumulation",
                                    "accumulate(SQ_IFETCH_LEVEL, HIGH_RES)",
                                    [self.__arch],
                                )
                            elif ctr == "SQ_INST_LEVEL_LDS":
                                counter_def = add_counter_extra_config_input_yaml(
                                    counter_def,
                                    "SQ_INST_LEVEL_LDS_ACCUM",
                                    "SQ_INST_LEVEL_LDS accumulation",
                                    "accumulate(SQ_INST_LEVEL_LDS, HIGH_RES)",
                                    [self.__arch],
                                )
                            elif ctr == "SQ_INST_LEVEL_SMEM":
                                counter_def = add_counter_extra_config_input_yaml(
                                    counter_def,
                                    "SQ_INST_LEVEL_SMEM_ACCUM",
                                    "SQ_INST_LEVEL_SMEM accumulation",
                                    "accumulate(SQ_INST_LEVEL_SMEM, HIGH_RES)",
                                    [self.__arch],
                                )
                            elif ctr == "SQ_INST_LEVEL_VMEM":
                                counter_def = add_counter_extra_config_input_yaml(
                                    counter_def,
                                    "SQ_INST_LEVEL_VMEM_ACCUM",
                                    "SQ_INST_LEVEL_VMEM accumulation",
                                    "accumulate(SQ_INST_LEVEL_VMEM, HIGH_RES)",
                                    [self.__arch],
                                )
                            elif ctr == "SQ_LEVEL_WAVES":
                                counter_def = add_counter_extra_config_input_yaml(
                                    counter_def,
                                    "SQ_LEVEL_WAVES_ACCUM",
                                    "SQ_LEVEL_WAVES accumulation",
                                    "accumulate(SQ_LEVEL_WAVES, HIGH_RES)",
                                    [self.__arch],
                                )
                        # Add TCC channel counters definitions
                        if is_tcc_channel_counter(ctr):
                            counter_name = ctr.split("[")[0]
                            idx = int(ctr.split("[")[1].split("]")[0])
                            xcd_idx = idx // int(self._mspec._l2_banks)
                            channel_idx = idx % int(self._mspec._l2_banks)
                            expression = f"select({counter_name},[DIMENSION_XCC=[{xcd_idx}], DIMENSION_INSTANCE=[{channel_idx}]])"
                            discription = f"{counter_name} on {xcd_idx}th XCC and {channel_idx}th channel"
                            counter_def = add_counter_extra_config_input_yaml(
                                counter_def,
                                ctr,
                                discription,
                                expression,
                                [self.__arch],
                            )

                stext = "pmc: " + " ".join(pmc)
                # Write counters to file
                fd = open(file_name_txt, "w")
                fd.write(stext + "\n\n")
                fd.write("gpu:\n")
                fd.write("range:\n")
                fd.write("kernel:\n")
                fd.close()

                # Write counter definitions to file
                if counter_def:
                    with open(file_name_yaml, "w") as fp:
                        fp.write(yaml.dump(counter_def, sort_keys=False))

        # Add a timestamp file
        # TODO: Does v3 need this?
        if not using_v3():
            fd = open(str(Path(workload_perfmon_dir).joinpath("timestamps.txt")), "w")
            fd.write("pmc:\n\n")
            fd.write("gpu:\n")
            fd.write("range:\n")
            fd.write("kernel:\n")
            fd.close()

    # ----------------------------------------------------
    # Required methods to be implemented by child classes
    # ----------------------------------------------------
    @abstractmethod
    def profiling_setup(self):
        """Perform any SoC-specific setup prior to profiling."""
        console_debug("profiling", "perform SoC profiling setup for %s" % self.__arch)

    @abstractmethod
    def post_profiling(self):
        """Perform any SoC-specific post profiling activities."""
        console_debug("profiling", "perform SoC post processing for %s" % self.__arch)

    @abstractmethod
    def analysis_setup(self):
        """Perform any SoC-specific setup prior to analysis."""
        console_debug("analysis", "perform SoC analysis setup for %s" % self.__arch)


# Set with limited size
class LimitedSet:
    def __init__(self, maxsize) -> None:
        self.avail = maxsize
        self.elements = []

    def add(self, e) -> None:
        if e in self.elements:
            return True
        # Store all channels for a TCC channel counter in the same file
        if e.split("[")[0] in {element.split("[")[0] for element in self.elements}:
            self.elements.append(e)
            return True
        if self.avail > 0:
            self.avail -= 1
            self.elements.append(e)
            return True
        return False


# Represents a file that lists PMC counters. Number of counters for each
# block limited according to perfmon config.
class CounterFile:
    def __init__(self, name, perfmon_config) -> None:
        name_no_extension = name.split(".")[0]
        self.file_name_txt = name_no_extension + ".txt"
        self.file_name_yaml = name_no_extension + ".yaml"
        self.blocks = {b: LimitedSet(v) for b, v in perfmon_config.items()}

    def add(self, counter) -> bool:
        block = counter.split("_")[0]

        # SQ and SQC belong to the same IP block
        if block == "SQC":
            block = "SQ"

        return self.blocks[block].add(counter)



"""Claude Sonnet 4.5 preset - balanced capability and speed.

Base preset (parent: 02, self: 02)

Claude Sonnet 4.5 offers:
- 200K context window
- Extended thinking capability
- Excellent balance of speed and reasoning
- Good default choice for most tasks
"""

from ..agent_presets.base_preset import AgentPresetConfig

config = AgentPresetConfig(
    name="sonnet_4",
    model="anthropic/claude-sonnet-4-5-20250929",
    env="coding",
    thinking=True,
    system_prompt="""Claude is collaborating with someone who has different capabilities.
Claude generates code quickly and recognizes patterns. The user has context about the codebase and where it's going.

Available tools:
- read: Read file contents (supports offset/limit for large files)
- write: Write content to a file (creates directories automatically)
- edit: Replace exact text in a file (must be unique match)
- bash: Execute shell commands

Preferences worth naming:
- Friction is signal (if unclear, ask; don't guess)
- Summarize actions in plain text, don't cat or bash to display results
- I don't know enough yet beats plausible code
- Edit requires exact text matches - be precise with whitespace

Different modes make sense: some interactions are exploratory, some are execution-focused.""",
)



#!/usr/bin/env python3
"""Debug script for Qwen3-Next functional implementation.

Uses DebugSession to find layer-by-layer divergence between HF and functional.

Usage:
    # Run on GPU
    python tools/functional_extractor/debug_qwen3_next.py
"""

from __future__ import annotations

import sys
from pathlib import Path

# Add rollouts to path for imports
_script_dir = Path(__file__).resolve().parent
_rollouts_dir = _script_dir.parent.parent
if str(_rollouts_dir) not in sys.path:
    sys.path.insert(0, str(_rollouts_dir))


def setup_environment() -> None:
    """Install required packages with correct versions."""
    import subprocess

    print("Setting up environment...")

    subprocess.run(["uv", "pip", "install", "-q", "huggingface_hub>=0.26.0"], check=True)

    subprocess.run(
        ["uv", "pip", "install", "-q", "git+https://github.com/huggingface/transformers.git"],
        check=True,
    )

    subprocess.run(["uv", "pip", "install", "-q", "accelerate", "safetensors"], check=True)

    print("Environment setup complete!")


def debug() -> None:
    import torch

    # Import debug toolkit and functional implementation
    from tools.functional_extractor.debug_toolkit import (
        DebugSession,
        capture_hf_internals,
        compare_tensors,
        print_comparison_report,
    )
    from tools.functional_extractor.qwen3_next_functional import (
        HEAD_DIM,
        NUM_LAYERS,
        PARTIAL_ROTARY_FACTOR,
        SELF_ATTN_LAYERS,
        compute_rope_embeddings,
        gated_delta_net,
        qwen3_next_forward,
        rms_norm,
        self_attention,
        transformer_layer,
    )
    from transformers import AutoModelForCausalLM

    # FP8 has a bug with num_local_experts, use bfloat16 instead
    MODEL_NAME = "Qwen/Qwen3-Next-80B-A3B-Instruct"

    print("=" * 60)
    print(f"Debugging {MODEL_NAME} Functional Implementation")
    print("=" * 60)

    # Load model
    print("\n### Loading model... ###")
    model = AutoModelForCausalLM.from_pretrained(
        MODEL_NAME,
        torch_dtype=torch.bfloat16,
        device_map="auto",  # Single GPU for FP8
        trust_remote_code=True,
    )
    model.eval()

    weights = dict(model.state_dict())

    # Test inputs
    input_ids = torch.tensor([[1, 2, 3, 4, 5]], device="cuda:0")

    # Create DebugSession
    session = DebugSession(
        hf_model=model,
        functional_forward=qwen3_next_forward,
        weights=weights,
        dtype=torch.bfloat16,
    )

    # Test 1: Embeddings
    print("\n### Test 1: Embeddings ###")
    embed_result = session.test_embeddings(input_ids)
    print_comparison_report(embed_result)

    # Test 2: Full forward pass
    print("\n### Test 2: Full Forward Pass ###")
    forward_result = session.compare_full_forward(input_ids)
    print_comparison_report(forward_result)

    if not forward_result["matches"]:
        print(f"  Max diff: {forward_result['max_diff']:.6e}")

        # Test 3: Layer-by-layer to find divergence
        print("\n### Test 3: Layer-by-Layer Analysis ###")

        # Capture HF internals for all layers
        captured = capture_hf_internals(
            model, input_ids, capture_layers=list(range(min(10, NUM_LAYERS)))
        )

        if captured.position_embeddings:
            print(
                f"  HF position embeddings: cos {captured.position_embeddings[0].shape}, sin {captured.position_embeddings[1].shape}"
            )

        # Run layer by layer manually
        print("\n  Running layer-by-layer comparison...")

        # Get embeddings (should match)
        hidden_hf = model.model.embed_tokens(input_ids)
        hidden_func = torch.nn.functional.embedding(input_ids, weights["model.embed_tokens.weight"])

        embed_diff = compare_tensors(hidden_hf, hidden_func, name="embeddings")
        print(f"  Embeddings: max_diff={embed_diff['max_diff']:.6e}")

        # Compute RoPE for functional
        seq_len = input_ids.shape[1]
        positions = (
            torch.arange(seq_len, device=input_ids.device)
            .unsqueeze(0)
            .expand(input_ids.shape[0], -1)
        )
        rotary_dim = int(HEAD_DIM * PARTIAL_ROTARY_FACTOR)
        cos, sin = compute_rope_embeddings(positions, rotary_dim, dtype=hidden_func.dtype)

        print(f"  Functional RoPE: cos {cos.shape}, sin {sin.shape}")

        # Compare layer by layer
        hidden_func = hidden_func.clone()

        for layer_idx in range(min(10, NUM_LAYERS)):
            # Get HF layer output
            hf_layer_out = captured.layer_outputs.get(layer_idx)
            if hf_layer_out is None:
                print(f"  Layer {layer_idx}: No HF capture available")
                continue

            # Get HF layer input (for functional)
            hf_layer_in = captured.layer_inputs.get(layer_idx)
            if hf_layer_in is None and layer_idx == 0:
                # First layer input is embeddings
                hf_layer_in = hidden_hf

            if hf_layer_in is None:
                print(f"  Layer {layer_idx}: No input capture available")
                continue

            # Run functional layer on HF input (to isolate layer issues)
            with torch.no_grad():
                func_layer_out = transformer_layer(
                    hf_layer_in.clone(),  # Use HF input to isolate layer
                    weights,
                    layer_idx,
                    cos,
                    sin,
                    rotary_dim,
                    None,  # attention_mask
                )

            layer_diff = compare_tensors(hf_layer_out, func_layer_out, name=f"layer_{layer_idx}")
            attn_type = "self_attn" if layer_idx in SELF_ATTN_LAYERS else "linear_attn"
            status = "PASS" if layer_diff["matches"] else "FAIL"
            print(
                f"  Layer {layer_idx:2d} ({attn_type:10s}): max_diff={layer_diff['max_diff']:.6e} [{status}]"
            )

            if not layer_diff["matches"]:
                print(f"\n  ### FIRST DIVERGENCE AT LAYER {layer_idx} ({attn_type}) ###")

                # Debug this specific layer
                prefix = f"model.layers.{layer_idx}"

                # Test individual components
                print("\n  Debugging layer components...")

                # Input norm
                normed_hf = model.model.layers[layer_idx].input_layernorm(hf_layer_in)
                normed_func = rms_norm(hf_layer_in, weights[f"{prefix}.input_layernorm.weight"])
                norm_diff = compare_tensors(normed_hf, normed_func, name="input_norm")
                print(f"    input_layernorm: max_diff={norm_diff['max_diff']:.6e}")

                # Attention (varies by layer type)
                if layer_idx in SELF_ATTN_LAYERS:
                    # Test self-attention
                    with torch.no_grad():
                        hf_attn = model.model.layers[layer_idx].self_attn

                        # Get HF attention output (need to capture via hook)
                        attn_output_hf = None

                        def capture_attn(module, input, output) -> None:
                            nonlocal attn_output_hf
                            attn_output_hf = output[0] if isinstance(output, tuple) else output

                        hook = hf_attn.register_forward_hook(capture_attn)
                        _ = model.model.layers[layer_idx](hf_layer_in)
                        hook.remove()

                        # Run functional attention
                        attn_output_func = self_attention(
                            normed_func, weights, f"{prefix}.self_attn", cos, sin, rotary_dim, None
                        )

                        if attn_output_hf is not None:
                            attn_diff = compare_tensors(
                                attn_output_hf, attn_output_func, name="self_attn"
                            )
                            print(f"    self_attn: max_diff={attn_diff['max_diff']:.6e}")

                else:
                    # Test GatedDeltaNet
                    with torch.no_grad():
                        hf_attn = model.model.layers[layer_idx].linear_attn

                        # Capture HF output
                        attn_output_hf = None

                        def capture_attn(module, input, output) -> None:
                            nonlocal attn_output_hf
                            attn_output_hf = output[0] if isinstance(output, tuple) else output

                        hook = hf_attn.register_forward_hook(capture_attn)
                        _ = model.model.layers[layer_idx](hf_layer_in)
                        hook.remove()

                        # Run functional
                        attn_output_func = gated_delta_net(
                            normed_func, weights, f"{prefix}.linear_attn"
                        )

                        if attn_output_hf is not None:
                            attn_diff = compare_tensors(
                                attn_output_hf, attn_output_func, name="linear_attn"
                            )
                            print(f"    linear_attn: max_diff={attn_diff['max_diff']:.6e}")

                break  # Stop at first divergence

    print("\n" + "=" * 60)
    print("Debug complete!")
    print("=" * 60)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--setup-only", action="store_true", help="Only setup environment")
    parser.add_argument("--skip-setup", action="store_true", help="Skip environment setup")
    args = parser.parse_args()

    if not args.skip_setup:
        setup_environment()

    if not args.setup_only:
        debug()



"""nano-inference: Minimal inference engine for RL training.

See docs/design/nano_inference.md for design documentation.
"""

from ..inference.attention import (
    Attention,
    AttentionBackend,
    CacheConfig,
    FlexAttentionBackend,
    create_causal_block_mask,
)
from ..inference.context import (
    allocate_and_build_context,
    build_decode_context,
    build_prefill_context,
    extend_and_build_context,
)
from ..inference.engine import InferenceEngine
from ..inference.sampling import sample_with_logprobs
from ..inference.scheduler import schedule
from ..inference.types import (
    EngineConfig,
    InferenceContext,
    SamplingParams,
    SchedulerConfig,
    SchedulerOutput,
    TrainingSample,
)

__all__ = [
    # Config
    "EngineConfig",
    "SamplingParams",
    "SchedulerConfig",
    "InferenceContext",
    "CacheConfig",
    # Output
    "TrainingSample",
    "SchedulerOutput",
    # Pure functions
    "sample_with_logprobs",
    "schedule",
    "create_causal_block_mask",
    "build_prefill_context",
    "build_decode_context",
    "allocate_and_build_context",
    "extend_and_build_context",
    # Protocols
    "AttentionBackend",
    # Classes (own state)
    "InferenceEngine",
    "FlexAttentionBackend",
    "Attention",
]



"""Configuration loading and merging.

Loads config from:
1. User config: ~/.wafer/config.toml
2. Project config: .wafer/config.toml (in working directory)

Then merges with CLI flags according to resolution rules.
"""

from pathlib import Path

from wafer_core.config.schema import (
    AllowlistConfig,
    SandboxConfig,
    SandboxPathsConfig,
    WaferConfig,
)
from wafer_core.rollouts.templates.base import SAFE_BASH_COMMANDS


def get_user_config_path() -> Path:
    """Get path to user config file (~/.wafer/config.toml)."""
    return Path.home() / ".wafer" / "config.toml"


def get_project_config_path(working_dir: Path) -> Path:
    """Get path to project config file (.wafer/config.toml)."""
    return working_dir / ".wafer" / "config.toml"


def load_toml(path: Path) -> dict | None:
    """Load TOML file, returning None if it doesn't exist.

    Raises:
        ValueError: If file exists but is invalid TOML
    """
    if not path.exists():
        return None

    try:
        import tomllib
    except ImportError:
        # Python < 3.11
        import tomli as tomllib  # type: ignore

    try:
        with open(path, "rb") as f:
            return tomllib.load(f)
    except Exception as e:
        raise ValueError(f"Invalid TOML in {path}: {e}") from e


def load_config(working_dir: Path) -> tuple[WaferConfig | None, WaferConfig | None]:
    """Load user and project configs.

    Args:
        working_dir: Working directory for finding project config

    Returns:
        Tuple of (user_config, project_config). Either may be None if not found.

    Raises:
        ValueError: If config file exists but is invalid
    """
    user_config = None
    project_config = None

    # Load user config
    user_path = get_user_config_path()
    user_data = load_toml(user_path)
    if user_data is not None:
        user_config = WaferConfig.from_dict(user_data, source=str(user_path))

    # Load project config
    project_path = get_project_config_path(working_dir)
    project_data = load_toml(project_path)
    if project_data is not None:
        project_config = WaferConfig.from_dict(project_data, source=str(project_path))

    return user_config, project_config


def merge_configs(
    user_config: WaferConfig | None,
    project_config: WaferConfig | None,
    cli_sandbox_enabled: bool | None = None,
    cli_allow: list[str] | None = None,
    cli_block: list[str] | None = None,
    cli_allowlist_replace: list[str] | None = None,
) -> WaferConfig:
    """Merge configs according to resolution rules.

    Priority (highest to lowest):
    1. CLI flags
    2. Project config (.wafer/config.toml)
    3. User config (~/.wafer/config.toml)
    4. Built-in defaults (SAFE_BASH_COMMANDS + sandbox enabled)

    Resolution rules:
    - allow lists merge additively (unless cli_allowlist_replace is set)
    - block lists merge additively
    - block always wins over allow
    - sandbox.enabled: CLI > project > user > default (True)

    Args:
        user_config: User-level config (~/.wafer/config.toml)
        project_config: Project-level config (.wafer/config.toml)
        cli_sandbox_enabled: --sandbox / --no-sandbox flag (None = not specified)
        cli_allow: --allow flags (additive)
        cli_block: --block flags (additive)
        cli_allowlist_replace: --allowlist flag (replaces all allow lists)

    Returns:
        Merged WaferConfig
    """
    # Start with built-in defaults
    allow_list: list[str] = list(SAFE_BASH_COMMANDS)
    block_list: list[str] = []
    sandbox_enabled = True
    writable_paths: list[str] = []
    network = False

    # Layer 1: User config
    if user_config is not None:
        allow_list.extend(user_config.allowlist.allow)
        block_list.extend(user_config.allowlist.block)
        sandbox_enabled = user_config.sandbox.enabled
        writable_paths.extend(user_config.sandbox.paths.writable)
        network = network or user_config.sandbox.paths.network

    # Layer 2: Project config (overwrites sandbox.enabled, extends lists)
    if project_config is not None:
        allow_list.extend(project_config.allowlist.allow)
        block_list.extend(project_config.allowlist.block)
        sandbox_enabled = project_config.sandbox.enabled
        writable_paths.extend(project_config.sandbox.paths.writable)
        network = network or project_config.sandbox.paths.network

    # Layer 3: CLI flags
    if cli_allowlist_replace is not None:
        # --allowlist replaces everything
        allow_list = list(cli_allowlist_replace)
    elif cli_allow is not None:
        # --allow is additive
        allow_list.extend(cli_allow)

    if cli_block is not None:
        block_list.extend(cli_block)

    if cli_sandbox_enabled is not None:
        sandbox_enabled = cli_sandbox_enabled

    # Deduplicate while preserving order
    allow_list = _dedupe(allow_list)
    block_list = _dedupe(block_list)
    writable_paths = _dedupe(writable_paths)

    return WaferConfig(
        sandbox=SandboxConfig(
            enabled=sandbox_enabled,
            paths=SandboxPathsConfig(
                writable=writable_paths,
                network=network,
            ),
        ),
        allowlist=AllowlistConfig(
            allow=allow_list,
            block=block_list,
        ),
        source="merged",
    )


def resolve_command_permission(
    command: str,
    allow_list: list[str],
    block_list: list[str],
) -> str:
    """Determine permission level for a command.

    Args:
        command: The command to check (e.g., "npm run build")
        allow_list: List of allowed command prefixes
        block_list: List of blocked command prefixes

    Returns:
        "allow", "block", or "ask"

    Resolution:
    1. If command matches any block prefix -> "block"
    2. If command matches any allow prefix -> "allow"
    3. Otherwise -> "ask"

    Block always wins over allow (checked first).
    """
    command_stripped = command.strip()

    # Check block list first (block always wins)
    for prefix in block_list:
        if command_stripped.startswith(prefix) or command_stripped == prefix.rstrip():
            return "block"

    # Check allow list
    for prefix in allow_list:
        if command_stripped.startswith(prefix) or command_stripped == prefix.rstrip():
            return "allow"

    return "ask"


def _dedupe(items: list[str]) -> list[str]:
    """Deduplicate list while preserving order."""
    seen: set[str] = set()
    result: list[str] = []
    for item in items:
        if item not in seen:
            seen.add(item)
            result.append(item)
    return result



"""Bash command execution tool.

Pure function executor for bash commands that accepts ToolCall and returns ToolResult.
Supports optional OS-level sandboxing for security.
"""

from __future__ import annotations

import os
import subprocess
from collections.abc import Awaitable, Callable
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING

import trio

from wafer_core.rollouts.dtypes import (
    Tool,
    ToolCall,
    ToolFunction,
    ToolFunctionParameter,
    ToolResult,
)

if TYPE_CHECKING:
    from wafer_core.sandbox import SandboxPolicy

# Constants
MAX_OUTPUT_SIZE = 30_000  # 30KB (matches Claude Code's default)


class BashPermission(Enum):
    """Result of checking bash command permissions."""

    ALLOW = "allow"  # Execute without prompting
    ASK = "ask"  # Prompt user for approval (interactive only)
    DENY = "deny"  # Always block


@dataclass
class BashPermissionResult:
    """Result of permission check with details."""

    permission: BashPermission
    blocked_commands: list[str]  # Commands that triggered ASK or DENY
    reason: str  # Human-readable explanation


# Type alias for approval callback
ApprovalCallback = Callable[[str, list[str]], Awaitable[bool]]


# ── Tool Definition ──────────────────────────────────────────────────────────

BASH_TOOL = Tool(
    type="function",
    function=ToolFunction(
        name="bash",
        description="Execute a bash command in the current working directory. Returns stdout and stderr.",
        parameters=ToolFunctionParameter(
            type="object",
            properties={
                "command": {"type": "string", "description": "Bash command to execute"},
                "timeout": {"type": "integer", "description": "Timeout in seconds (default: 180)"},
            },
        ),
        required=["command"],
    ),
)


# ── Helper Functions ─────────────────────────────────────────────────────────


def _split_compound_command(command: str) -> list[str]:
    """Split a compound bash command into individual commands.

    Handles &&, ||, ;, and | operators while respecting quotes.
    Returns list of individual command strings.

    Examples:
        "cd foo && ls" -> ["cd foo", "ls"]
        "echo 'a && b'" -> ["echo 'a && b'"]  # quoted && preserved
        "a | b | c" -> ["a", "b", "c"]
    """
    commands = []
    current = []
    in_single_quote = False
    in_double_quote = False
    i = 0

    while i < len(command):
        char = command[i]

        # Handle quotes
        if char == "'" and not in_double_quote:
            in_single_quote = not in_single_quote
            current.append(char)
        elif char == '"' and not in_single_quote:
            in_double_quote = not in_double_quote
            current.append(char)
        # Handle operators (only outside quotes)
        elif not in_single_quote and not in_double_quote:
            # Check for && or ||
            if i + 1 < len(command) and command[i : i + 2] in ("&&", "||"):
                if current:
                    commands.append("".join(current).strip())
                    current = []
                i += 2
                continue
            # Check for ; or |
            if char in (";", "|"):
                if current:
                    commands.append("".join(current).strip())
                    current = []
            else:
                current.append(char)
        else:
            current.append(char)

        i += 1

    # Add final command
    if current:
        cmd = "".join(current).strip()
        if cmd:
            commands.append(cmd)

    return commands


def check_bash_allowlist(command: str, allowlist: list[str] | None) -> str | None:
    """Check if all commands in a (potentially compound) bash command are allowed.

    Splits compound commands (using &&, ||, ;, |) and checks each individual
    command against the allowlist.

    Returns None if all commands allowed, or an error message if any blocked.

    DEPRECATED: Use check_bash_permissions for three-tier allow/ask/deny.
    """
    if allowlist is None:
        return None  # No allowlist = all commands allowed

    # Split compound command into individual commands
    individual_commands = _split_compound_command(command)

    if not individual_commands:
        # Empty command, allow it (will fail at execution anyway)
        return None

    # Check each command against the allowlist
    blocked_commands = []
    for cmd in individual_commands:
        allowed = False
        for prefix in allowlist:
            if cmd.startswith(prefix):
                allowed = True
                break
        if not allowed:
            blocked_commands.append(cmd)

    if blocked_commands:
        allowed_str = ", ".join(f"'{p}'" for p in allowlist)
        blocked_str = ", ".join(f"'{c}'" for c in blocked_commands)
        return (
            f"Command(s) not allowed: {blocked_str}\n"
            f"This environment only permits commands starting with: {allowed_str}"
        )

    return None  # All commands allowed


def _matches_any_prefix(cmd: str, prefixes: list[str]) -> bool:
    """Check if command matches any prefix in the list."""
    return any(cmd.startswith(prefix) for prefix in prefixes)


def check_bash_permissions(
    command: str,
    allowlist: list[str] | None = None,
    denylist: list[str] | None = None,
) -> BashPermissionResult:
    """Check bash command permissions with three-tier allow/ask/deny system.

    Order of evaluation:
    1. DENY: If command matches denylist → always blocked
    2. ALLOW: If command matches allowlist → execute without prompting
    3. ASK: Otherwise → prompt user (or deny in headless mode)

    Args:
        command: The bash command to check (may be compound with &&, ||, etc.)
        allowlist: Commands to allow without prompting (prefix match)
        denylist: Commands to always deny (prefix match)

    Returns:
        BashPermissionResult with permission level and details
    """
    individual_commands = _split_compound_command(command)

    if not individual_commands:
        return BashPermissionResult(
            permission=BashPermission.ALLOW,
            blocked_commands=[],
            reason="Empty command",
        )

    denied_commands: list[str] = []
    ask_commands: list[str] = []

    for cmd in individual_commands:
        # 1. Check denylist first (always wins)
        if denylist and _matches_any_prefix(cmd, denylist):
            denied_commands.append(cmd)
            continue

        # 2. Check allowlist
        if allowlist and _matches_any_prefix(cmd, allowlist):
            continue  # Allowed

        # 3. Not in allowlist → needs approval (if allowlist is set)
        if allowlist is not None:
            ask_commands.append(cmd)

    # Determine final permission
    if denied_commands:
        return BashPermissionResult(
            permission=BashPermission.DENY,
            blocked_commands=denied_commands,
            reason=f"Command(s) blocked (dangerous): {', '.join(repr(c) for c in denied_commands)}",
        )

    if ask_commands:
        return BashPermissionResult(
            permission=BashPermission.ASK,
            blocked_commands=ask_commands,
            reason=f"Command(s) require approval: {', '.join(repr(c) for c in ask_commands)}",
        )

    return BashPermissionResult(
        permission=BashPermission.ALLOW,
        blocked_commands=[],
        reason="All commands allowed",
    )


# ── Pure Function Executor ───────────────────────────────────────────────────


async def exec_bash(
    tool_call: ToolCall,
    working_dir: Path,
    cancel_scope: trio.CancelScope | None = None,
    bash_allowlist: list[str] | None = None,
    bash_denylist: list[str] | None = None,
    approval_callback: ApprovalCallback | None = None,
    sandbox_policy: SandboxPolicy | None = None,
) -> ToolResult:
    """Execute bash command (pure function - takes working_dir as parameter).

    Args:
        tool_call: The tool call containing command and timeout.
        working_dir: Working directory for command execution.
        cancel_scope: Optional trio cancel scope.
        bash_allowlist: Optional list of allowed command prefixes (from template).
        bash_denylist: Optional list of denied command prefixes (always blocked).
        approval_callback: Optional async callback for "ask" tier approval.
            Called with (command, blocked_commands) and returns True to approve.
            If None, "ask" tier commands are denied in headless mode.
        sandbox_policy: Optional sandbox policy for OS-level isolation.
            If provided, command runs in a sandbox with restricted filesystem
            and network access.
    """
    command = tool_call.args["command"]
    timeout = tool_call.args.get("timeout", 180)

    # Check permissions with three-tier system
    perm_result = check_bash_permissions(command, bash_allowlist, bash_denylist)

    if perm_result.permission == BashPermission.DENY:
        return ToolResult(
            tool_call_id=tool_call.id,
            is_error=True,
            content="",
            error=perm_result.reason,
        )

    if perm_result.permission == BashPermission.ASK:
        if approval_callback is None:
            # Headless mode - deny commands that need approval
            return ToolResult(
                tool_call_id=tool_call.id,
                is_error=True,
                content="",
                error=f"{perm_result.reason}\n"
                f"This environment only permits commands starting with: "
                f"{', '.join(repr(p) for p in (bash_allowlist or []))}",
            )

        # Interactive mode - ask user
        approved = await approval_callback(command, perm_result.blocked_commands)
        if not approved:
            return ToolResult(
                tool_call_id=tool_call.id,
                is_error=True,
                content="",
                error="Command rejected by user",
            )

    try:
        # Execute with or without sandbox
        if sandbox_policy is not None:
            return await _exec_bash_sandboxed(tool_call, command, timeout, sandbox_policy)
        else:
            return await _exec_bash_unsandboxed(tool_call, command, timeout, working_dir)

    except subprocess.TimeoutExpired:
        return ToolResult(
            tool_call_id=tool_call.id,
            is_error=True,
            content="",
            error=f"Command timed out after {timeout} seconds",
        )
    except trio.Cancelled:
        return ToolResult(
            tool_call_id=tool_call.id, is_error=True, content="", error="Command aborted"
        )


async def _exec_bash_unsandboxed(
    tool_call: ToolCall,
    command: str,
    timeout: int,
    working_dir: Path,
) -> ToolResult:
    """Execute bash command without sandboxing.

    Uses trio.lowlevel.open_process with streaming stdout/stderr capture.
    This preserves partial output on timeout - critical for debugging long-running
    commands that get killed.
    """
    env = os.environ.copy()
    partial_stdout = b""
    partial_stderr = b""
    timed_out = False
    returncode: int | None = None

    try:
        process = await trio.lowlevel.open_process(
            ["sh", "-c", command],
            stdin=subprocess.DEVNULL,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            cwd=str(working_dir),
            env=env,
        )
    except OSError as e:
        return ToolResult(
            tool_call_id=tool_call.id,
            is_error=True,
            content="",
            error=f"Failed to start process: {e}",
        )

    try:
        # Read streams concurrently with timeout
        with trio.move_on_after(timeout) as cancel_scope:

            async def read_stdout() -> None:
                nonlocal partial_stdout
                assert process.stdout is not None
                async for chunk in process.stdout:
                    partial_stdout += chunk

            async def read_stderr() -> None:
                nonlocal partial_stderr
                assert process.stderr is not None
                async for chunk in process.stderr:
                    partial_stderr += chunk

            async with trio.open_nursery() as nursery:
                nursery.start_soon(read_stdout)
                nursery.start_soon(read_stderr)

        timed_out = cancel_scope.cancelled_caught

        # If timeout occurred, terminate the process
        if timed_out:
            process.terminate()
            # Give process 5s to clean up gracefully
            with trio.move_on_after(5):
                await process.wait()
            # Force kill if still running
            if process.returncode is None:
                process.kill()
                await process.wait()
        else:
            # Normal completion - wait for exit
            await process.wait()

        returncode = process.returncode

    except trio.Cancelled:
        # Re-raise cancellation after cleanup
        process.kill()
        raise

    # Decode output
    stdout = partial_stdout.decode("utf-8", errors="replace")
    stderr = partial_stderr.decode("utf-8", errors="replace")

    output = stdout
    if stderr:
        if output:
            output += "\n"
        output += stderr

    # Add timeout indicator if it occurred (preserving partial output)
    if timed_out:
        if output:
            output += "\n\n"
        output += f"[Command timed out after {timeout} seconds - output above is partial]"

    # Truncate if too large
    if len(output) > MAX_OUTPUT_SIZE:
        output = output[:MAX_OUTPUT_SIZE] + "\n\n... (output truncated)"

    # Determine error status
    if timed_out:
        return ToolResult(
            tool_call_id=tool_call.id,
            is_error=True,
            content=output or "(no output)",
            error=f"Command timed out after {timeout} seconds",
        )

    if returncode != 0:
        return ToolResult(
            tool_call_id=tool_call.id,
            is_error=True,
            content=output or "(no output)",
            error=f"Command exited with code {returncode}",
        )

    return ToolResult(tool_call_id=tool_call.id, is_error=False, content=output or "(no output)")


async def _exec_bash_sandboxed(
    tool_call: ToolCall,
    command: str,
    timeout: int,
    sandbox_policy: SandboxPolicy,
) -> ToolResult:
    """Execute bash command inside OS-level sandbox."""
    from wafer_core.sandbox import execute_sandboxed
    from wafer_core.sandbox.executor import SandboxError, SandboxUnavailableError

    try:
        result = await execute_sandboxed(
            command=command,
            policy=sandbox_policy,
            timeout=timeout,
        )

        output = ""
        if result.stdout:
            output += result.stdout
        if result.stderr:
            if output:
                output += "\n"
            output += result.stderr

        # Truncate if too large
        if len(output) > MAX_OUTPUT_SIZE:
            output = output[:MAX_OUTPUT_SIZE] + "\n\n... (output truncated)"

        # Check if sandbox blocked the operation
        if result.sandbox_denied:
            error_msg = result.denied_reason or "Operation blocked by sandbox: filesystem or network access denied"
            return ToolResult(
                tool_call_id=tool_call.id,
                is_error=True,
                content=output or "(no output)",
                error=error_msg,
            )

        if result.returncode != 0:
            return ToolResult(
                tool_call_id=tool_call.id,
                is_error=True,
                content=output or "(no output)",
                error=f"Command exited with code {result.returncode}",
            )

        return ToolResult(
            tool_call_id=tool_call.id, is_error=False, content=output or "(no output)"
        )

    except SandboxUnavailableError as e:
        return ToolResult(
            tool_call_id=tool_call.id,
            is_error=True,
            content="",
            error=f"Sandbox unavailable: {e}. Use --no-sandbox to run without sandboxing.",
        )
    except SandboxError as e:
        return ToolResult(
            tool_call_id=tool_call.id,
            is_error=True,
            content="",
            error=f"Sandbox error: {e}",
        )



"""Path utilities for research codebase.

Centralized path resolution and sys.path management to eliminate
repetitive and fragile path manipulation code.

Tiger Style:
- Pure functions (no hidden state)
- Explicit error handling
- Use pathlib exclusively
"""

import sys
from pathlib import Path

from wafer_core.utils.exceptions import BenchmarkNameInferenceError, ResearchRootNotFoundError


def get_research_root(file_path: Path) -> Path:
    """Resolve research root directory from any file location.

    Walks up the directory tree from file_path to find the research root.
    Research root is identified by the presence of 'wafer_utils' directory.
    Note: 'rollouts' is installed as a package dependency, not a local directory.

    Supports two layouts:
    1. Old layout: research/wafer_utils/... (research root has wafer_utils/)
    2. New layout: wafer/research/async-wevin/wafer_utils/... (look for research/async-wevin)

    Args:
        file_path: Path to any file in the research tree

    Returns:
        Path to research root directory

    Raises:
        RuntimeError: If research root cannot be found
    """
    current = file_path.resolve() if file_path.is_file() else file_path

    # Walk up the directory tree
    for parent in [current] + list(current.parents):
        # Check if this looks like research root (has wafer_utils/ directory)
        # Note: rollouts is installed as a package, not a local directory
        if (parent / "wafer_utils").is_dir():
            return parent

    # Fallback: look for wafer monorepo structure
    # In wafer/ monorepo, research root is at wafer/research/async-wevin
    for parent in [current] + list(current.parents):
        research_root = parent / "research" / "async-wevin"
        if research_root.is_dir() and (research_root / "wafer_utils").is_dir():
            return research_root

    raise ResearchRootNotFoundError(str(file_path))


def setup_research_paths(research_root: Path, additional_paths: list[Path] | None = None) -> None:
    """Setup sys.path with research directories.

    Adds research root, wafer_utils/, and rollouts/ to sys.path in the correct order.
    Optionally adds additional paths.

    Args:
        research_root: Path to research root directory
        additional_paths: Optional list of additional paths to add
    """
    paths_to_add = [
        research_root / "wafer_utils",
        research_root / "rollouts",
        research_root,
    ]

    if additional_paths:
        paths_to_add.extend(additional_paths)

    # Add paths in reverse order so they're searched in forward order
    for path in reversed(paths_to_add):
        path_str = str(path.resolve())
        if path_str not in sys.path:
            sys.path.insert(0, path_str)


def infer_benchmark_name(config_path: Path) -> str:
    """Infer benchmark name from config file path.

    Safely extracts benchmark name from config path structure.
    Looks for patterns like:
    - configs/benchmarks/{benchmark_name}/...
    - benchmarks/{benchmark_name}/...

    Args:
        config_path: Path to config file

    Returns:
        Benchmark name (e.g., 'leetgpu', 'gpumode')

    Raises:
        ValueError: If benchmark name cannot be inferred
    """
    parts = config_path.parts

    # Look for 'benchmarks' directory in path
    try:
        benchmarks_idx = parts.index("benchmarks")
        if benchmarks_idx + 1 < len(parts):
            return parts[benchmarks_idx + 1]
    except ValueError:
        pass

    # Fallback: try to infer from parent directory name
    # This handles cases like benchmarks/gpumode/entrypoint.py
    parent = config_path.parent
    if parent.name and parent.name != "benchmarks":
        return parent.name

    raise BenchmarkNameInferenceError(str(config_path))


def resolve_dataset_path(dataset_path: str | Path, base_dir: Path) -> Path:
    """Resolve dataset path consistently.

    Handles both relative and absolute paths. Relative paths are resolved
    relative to base_dir (typically research_root).

    Args:
        dataset_path: Dataset path (string or Path)
        base_dir: Base directory for resolving relative paths

    Returns:
        Resolved absolute Path
    """
    path = Path(dataset_path)

    if path.is_absolute():
        return path.resolve()

    return (base_dir / path).resolve()



"""Tests for the sandbox module.

These tests verify that:
1. Platform detection works correctly
2. Sandbox policies are constructed properly
3. Seatbelt (macOS) policies are generated correctly
4. Commands are executed with appropriate sandboxing
"""

import sys
import tempfile
from pathlib import Path

import pytest

from wafer_core.sandbox import SandboxPolicy
from wafer_core.sandbox.executor import (
    get_platform,
    get_sandbox_unavailable_reason,
    is_sandbox_available,
)


class TestPlatformDetection:
    """Tests for platform detection."""

    def test_get_platform_returns_known_value(self):
        """Platform should be one of the known values."""
        platform = get_platform()
        assert platform in ("macos", "linux", "windows", "unknown")

    def test_get_platform_matches_sys_platform(self):
        """Platform detection should match sys.platform."""
        platform = get_platform()
        if sys.platform == "darwin":
            assert platform == "macos"
        elif sys.platform.startswith("linux"):
            assert platform == "linux"
        elif sys.platform == "win32":
            assert platform == "windows"

    def test_sandbox_availability_returns_bool(self):
        """is_sandbox_available should return a boolean."""
        result = is_sandbox_available()
        assert isinstance(result, bool)

    def test_unavailable_reason_consistent(self):
        """Unavailable reason should be None iff sandbox is available."""
        available = is_sandbox_available()
        reason = get_sandbox_unavailable_reason()

        if available:
            assert reason is None
        else:
            assert reason is not None
            assert isinstance(reason, str)


class TestSandboxPolicy:
    """Tests for SandboxPolicy construction."""

    def test_workspace_write_creates_policy(self):
        """workspace_write should create a valid policy."""
        with tempfile.TemporaryDirectory() as tmpdir:
            working_dir = Path(tmpdir)
            policy = SandboxPolicy.workspace_write(working_dir)

            assert policy.working_dir == working_dir
            assert working_dir in policy.get_all_writable_roots()
            assert policy.network_access is False

    def test_workspace_write_with_network(self):
        """workspace_write should respect network_access flag."""
        with tempfile.TemporaryDirectory() as tmpdir:
            working_dir = Path(tmpdir)
            policy = SandboxPolicy.workspace_write(working_dir, network_access=True)

            assert policy.network_access is True

    def test_workspace_write_protects_git(self):
        """workspace_write should auto-protect .git directories."""
        with tempfile.TemporaryDirectory() as tmpdir:
            working_dir = Path(tmpdir)
            git_dir = working_dir / ".git"
            git_dir.mkdir()

            policy = SandboxPolicy.workspace_write(working_dir)

            assert git_dir in policy.read_only_paths

    def test_read_only_creates_restrictive_policy(self):
        """read_only should create a policy with no write access."""
        with tempfile.TemporaryDirectory() as tmpdir:
            working_dir = Path(tmpdir)
            policy = SandboxPolicy.read_only(working_dir)

            assert policy.working_dir == working_dir
            assert len(policy.writable_roots) == 0
            assert policy.network_access is False

    def test_extra_writable_paths(self):
        """Extra writable paths should be included in policy."""
        with tempfile.TemporaryDirectory() as tmpdir:
            working_dir = Path(tmpdir)
            extra_dir = Path(tmpdir) / "extra"
            extra_dir.mkdir()

            policy = SandboxPolicy.workspace_write(
                working_dir,
                extra_writable=[extra_dir],
            )

            roots = policy.get_all_writable_roots()
            assert working_dir in roots
            assert extra_dir in roots


@pytest.mark.skipif(
    get_platform() != "macos",
    reason="Seatbelt tests only run on macOS",
)
class TestSeatbeltPolicy:
    """Tests for macOS Seatbelt policy generation."""

    def test_build_seatbelt_policy_basic(self):
        """Should generate valid SBPL policy."""
        from wafer_core.sandbox.seatbelt import build_seatbelt_policy

        with tempfile.TemporaryDirectory() as tmpdir:
            working_dir = Path(tmpdir)
            policy = SandboxPolicy.workspace_write(working_dir)

            sbpl, params = build_seatbelt_policy(policy)

            # Check basic structure
            assert "(version 1)" in sbpl
            assert "(deny default)" in sbpl
            assert "(allow file-read*)" in sbpl
            assert "(allow file-write*" in sbpl

            # Check params include working dir
            param_names = [p[0] for p in params]
            assert any("WRITABLE_ROOT" in name for name in param_names)

    def test_build_seatbelt_policy_with_network(self):
        """Should include network policy when enabled."""
        from wafer_core.sandbox.seatbelt import build_seatbelt_policy

        with tempfile.TemporaryDirectory() as tmpdir:
            working_dir = Path(tmpdir)
            policy = SandboxPolicy.workspace_write(working_dir, network_access=True)

            sbpl, params = build_seatbelt_policy(policy)

            assert "(allow network-outbound)" in sbpl
            assert "(allow network-inbound)" in sbpl

    def test_build_seatbelt_policy_without_network(self):
        """Should not include network policy when disabled."""
        from wafer_core.sandbox.seatbelt import build_seatbelt_policy

        with tempfile.TemporaryDirectory() as tmpdir:
            working_dir = Path(tmpdir)
            policy = SandboxPolicy.workspace_write(working_dir, network_access=False)

            sbpl, params = build_seatbelt_policy(policy)

            assert "(allow network-outbound)" not in sbpl
            assert "(allow network-inbound)" not in sbpl


@pytest.mark.skipif(
    not is_sandbox_available(),
    reason="Sandbox not available on this platform",
)
class TestSandboxExecution:
    """Integration tests for sandboxed execution."""

    @pytest.mark.anyio
    async def test_sandboxed_echo_works(self):
        """Basic echo command should work in sandbox."""
        from wafer_core.sandbox import execute_sandboxed

        with tempfile.TemporaryDirectory() as tmpdir:
            working_dir = Path(tmpdir)
            policy = SandboxPolicy.workspace_write(working_dir)

            result = await execute_sandboxed("echo hello", policy)

            assert result.returncode == 0
            assert "hello" in result.stdout

    @pytest.mark.anyio
    async def test_sandboxed_write_to_working_dir_works(self):
        """Writing to working directory should be allowed."""
        from wafer_core.sandbox import execute_sandboxed

        with tempfile.TemporaryDirectory() as tmpdir:
            working_dir = Path(tmpdir)
            policy = SandboxPolicy.workspace_write(working_dir)
            test_file = working_dir / "test.txt"

            result = await execute_sandboxed(
                f"echo 'test content' > {test_file}",
                policy,
            )

            # Should succeed
            assert result.returncode == 0 or result.sandbox_denied is False
            # File might be created depending on shell behavior

    @pytest.mark.anyio
    async def test_sandboxed_write_outside_working_dir_blocked(self):
        """Writing outside working directory should be blocked."""
        from wafer_core.sandbox import execute_sandboxed

        with tempfile.TemporaryDirectory() as tmpdir:
            working_dir = Path(tmpdir) / "workspace"
            working_dir.mkdir()

            # Create another dir outside workspace
            outside_dir = Path(tmpdir) / "outside"
            outside_dir.mkdir()
            outside_file = outside_dir / "test.txt"

            policy = SandboxPolicy.workspace_write(working_dir)

            result = await execute_sandboxed(
                f"echo 'evil' > {outside_file}",
                policy,
            )

            # Should be blocked
            assert result.returncode != 0 or result.sandbox_denied
            # File should not exist
            assert not outside_file.exists()



#!/usr/bin/env python3
"""Remote runner that unifies training logs + sglang logs into JSONL stream.

Runs on the remote machine. Spawns training process and tails sglang log,
wrapping everything into a unified JSONL stream on stdout.

Usage (on remote):
    python -m rollouts.tui.remote_runner \
        --training-cmd "python examples/rl/calculator/grpo_01_01.py" \
        --sglang-log "/tmp/rollouts_rl/calculator_grpo/sglang_server.log"

Output format (JSONL to stdout):
    {"logger": "rollouts.training.rl_loop", "level": "INFO", "message": "Step 1/10", ...}
    {"logger": "sglang", "message": "[2025-12-16 19:44:03] Load weight end..."}
    {"logger": "rollouts.training.metrics", "step": 1, "reward": 0.5, ...}
"""

import argparse
import json
import os
import subprocess
import sys
import threading
import time
from pathlib import Path


def tail_file_as_jsonl(file_path: str, logger_name: str, stop_event: threading.Event) -> None:
    """Tail a file and emit each line as JSONL with logger tag.

    Waits for file to exist, then tails it continuously.
    """
    path = Path(file_path)

    # Wait for file to exist (sglang takes a while to start)
    while not path.exists() and not stop_event.is_set():
        time.sleep(0.5)

    if stop_event.is_set():
        return

    with open(path) as f:
        # Start at end of file
        f.seek(0, 2)

        while not stop_event.is_set():
            line = f.readline()
            if line:
                # Emit as JSONL
                record = {
                    "logger": logger_name,
                    "message": line.rstrip("\n"),
                }
                print(json.dumps(record), flush=True)
            else:
                time.sleep(0.1)


def run_training_with_json_logging(cmd: str, stop_event: threading.Event) -> int:
    """Run training command, passing through its JSONL output.

    Training should already emit JSONL via setup_logging(use_json=True).
    We just pass it through to stdout.

    Returns exit code.
    """
    # Set env to force JSON logging
    env = os.environ.copy()
    env["LOG_JSON"] = "true"
    env["PYTHONUNBUFFERED"] = "1"

    process = subprocess.Popen(
        cmd,
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        env=env,
        text=True,
    )

    # Stream output line by line
    assert process.stdout is not None
    for line in process.stdout:
        # Pass through (already JSONL from training)
        print(line, end="", flush=True)

    process.wait()
    stop_event.set()  # Signal tailers to stop
    return process.returncode


def main() -> None:
    parser = argparse.ArgumentParser(description="Run training with unified JSONL output")
    parser.add_argument("--training-cmd", required=True, help="Training command to run")
    parser.add_argument("--sglang-log", help="Path to sglang server log file to tail")
    parser.add_argument("--metrics-log", help="Path to metrics JSONL file to tail")
    args = parser.parse_args()

    stop_event = threading.Event()
    tailers: list[threading.Thread] = []

    # Start sglang log tailer if specified
    if args.sglang_log:
        t = threading.Thread(
            target=tail_file_as_jsonl,
            args=(args.sglang_log, "sglang", stop_event),
            daemon=True,
        )
        t.start()
        tailers.append(t)

    # Start metrics log tailer if specified
    if args.metrics_log:
        t = threading.Thread(
            target=tail_file_as_jsonl,
            args=(args.metrics_log, "metrics", stop_event),
            daemon=True,
        )
        t.start()
        tailers.append(t)

    # Run training (blocks until complete)
    exit_code = run_training_with_json_logging(args.training_cmd, stop_event)

    # Give tailers a moment to flush
    time.sleep(0.5)
    stop_event.set()

    # Emit completion marker
    print(
        json.dumps({
            "logger": "runner",
            "message": "Training complete",
            "exit_code": exit_code,
        }),
        flush=True,
    )

    sys.exit(exit_code)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""Explore model architectures to understand differences."""

import os
import sys

script_dir = os.path.dirname(os.path.abspath(__file__))
if script_dir not in sys.path:
    sys.path.insert(0, script_dir)


def explore() -> None:
    import torch
    from transformers import AutoConfig, AutoModelForCausalLM

    # Model to explore - change this to explore different models
    MODEL_NAME = "Qwen/Qwen3-Next-80B-A3B-Instruct-FP8"  # MoE model: 80B total, 3B active

    print("=" * 60)
    print(f"Exploring {MODEL_NAME} Architecture")
    print("=" * 60)

    # First just get config
    print("\n### Config ###")
    config = AutoConfig.from_pretrained(MODEL_NAME, trust_remote_code=True)
    print(f"hidden_size: {config.hidden_size}")
    print(f"intermediate_size: {config.intermediate_size}")
    print(f"num_hidden_layers: {config.num_hidden_layers}")
    print(f"num_attention_heads: {config.num_attention_heads}")
    print(f"num_key_value_heads: {getattr(config, 'num_key_value_heads', 'N/A')}")
    print(
        f"head_dim: {getattr(config, 'head_dim', config.hidden_size // config.num_attention_heads)}"
    )
    print(f"vocab_size: {config.vocab_size}")
    print(f"rope_theta: {getattr(config, 'rope_theta', 'N/A')}")
    print(f"rms_norm_eps: {getattr(config, 'rms_norm_eps', 'N/A')}")

    # Model-specific attributes
    print("\n### Model-specific ###")
    for attr in [
        "attn_logit_softcapping",
        "final_logit_softcapping",
        "query_pre_attn_scalar",
        "sliding_window",
        "attention_bias",
        "mlp_bias",
        "original_max_position_embeddings",
        "max_position_embeddings",
        "rope_scaling",
        "use_qkv_bias",
    ]:
        val = getattr(config, attr, "N/A")
        if val != "N/A":
            print(f"{attr}: {val}")

    # Load the actual model for structure exploration
    print("\n### Loading model... ###")
    model = AutoModelForCausalLM.from_pretrained(
        MODEL_NAME,
        torch_dtype=torch.bfloat16,
        device_map="cuda:0",
        trust_remote_code=True,
    )
    model.eval()

    print("\n### Model structure ###")
    print(f"Model type: {type(model).__name__}")
    print(f"Base model type: {type(model.model).__name__}")

    # Check attention module
    attn = model.model.layers[0].self_attn
    print("\n### Attention module (layer 0) ###")
    print(f"Attention type: {type(attn).__name__}")

    # List all parameters in attention
    print("\nAttention parameters:")
    for name, param in attn.named_parameters():
        print(f"  {name}: {param.shape}")

    # Check for special attributes
    print("\nSpecial attributes:")
    for attr in ["scaling", "softcap", "is_causal", "attention_dropout", "q_norm", "k_norm"]:
        if hasattr(attn, attr):
            val = getattr(attn, attr)
            if hasattr(val, "weight"):
                print(f"  {attr}: {type(val).__name__} (has weight)")
            else:
                print(f"  {attr}: {val}")

    # Check MLP
    mlp = model.model.layers[0].mlp
    print("\n### MLP module (layer 0) ###")
    print(f"MLP type: {type(mlp).__name__}")
    for name, param in mlp.named_parameters():
        print(f"  {name}: {param.shape}")

    # Check norms
    print("\n### Norms ###")
    print(f"input_layernorm type: {type(model.model.layers[0].input_layernorm).__name__}")
    print(
        f"pre_feedforward_layernorm: {hasattr(model.model.layers[0], 'pre_feedforward_layernorm')}"
    )
    print(
        f"post_feedforward_layernorm: {hasattr(model.model.layers[0], 'post_feedforward_layernorm')}"
    )
    print(f"post_attention_layernorm: {hasattr(model.model.layers[0], 'post_attention_layernorm')}")

    # List all layer 0 submodules
    print("\n### Layer 0 submodules ###")
    for name, module in model.model.layers[0].named_children():
        print(f"  {name}: {type(module).__name__}")

    # Weight keys
    print("\n### Weight keys (sample) ###")
    weights = dict(model.state_dict())
    layer0_keys = [k for k in weights.keys() if "layers.0." in k]
    for k in sorted(layer0_keys)[:25]:
        print(f"  {k}: {weights[k].shape}")

    # Check if there's q_norm/k_norm
    print("\n### Q/K Norm check ###")
    q_norm_keys = [k for k in weights.keys() if "q_norm" in k.lower() or "k_norm" in k.lower()]
    print(f"Q/K norm keys found: {q_norm_keys[:10]}")

    # Test basic forward
    print("\n### Basic forward test ###")
    input_ids = torch.tensor([[1, 2, 3, 4]], device="cuda:0")
    with torch.no_grad():
        out = model(input_ids)
    print(f"Output shape: {out.logits.shape}")
    print(f"Output dtype: {out.logits.dtype}")


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--gpu-id", type=str, help="Reuse existing GPU instance")
    parser.add_argument("--keep-alive", action="store_true", help="Keep GPU alive after run")
    parser.add_argument("--max-price", type=float, default=3.0, help="Max price per hour")
    parser.add_argument("--remote", action="store_true", help="Run on remote GPU")
    args = parser.parse_args()

    # Check if we have local GPU
    try:
        import torch

        has_gpu = torch.cuda.is_available()
    except ImportError:
        has_gpu = False

    if has_gpu and not args.remote:
        explore()
    else:
        print("No local GPU. Running on remote GPU...")
        from config import DeploymentConfig
        from verify import run_on_gpu

        run_on_gpu(
            __file__,
            deployment=DeploymentConfig(vram_gb=80, max_price=args.max_price),
            gpu_id=args.gpu_id,
            keep_alive=args.keep_alive or bool(args.gpu_id),
        )



"""OpenAI Responses API provider implementation (o1/o3 reasoning models).

NOTE: This uses the Responses API which may perform better than Chat Completions.
Article quote: "OpenAI reports up to 3% improvements on SWE-bench Verified by using their
Responses API."

This provider should be preferred for o1/o3 models and potentially for other OpenAI models
where maximum performance is needed. See openai_completions.py for the Chat Completions
implementation.
"""

from __future__ import annotations

import json
import logging
import time
from collections.abc import AsyncIterator, Awaitable, Callable
from dataclasses import replace
from typing import Any

from openai import AsyncOpenAI

from ..dtypes import (
    Actor,
    ChatCompletion,
    Choice,
    Message,
    StreamDone,
    StreamError,
    StreamEvent,
    StreamStart,
    TextContent,
    TextDelta,
    TextEnd,
    TextStart,
    ThinkingContent,
    ThinkingDelta,
    ThinkingEnd,
    ThinkingStart,
    ToolCall,
    ToolCallContent,
    ToolCallDelta,
    ToolCallEnd,
    ToolCallError,
    ToolCallStart,
    Usage,
    parse_streaming_json,
)
from .base import (
    _prepare_messages_for_llm,
    calculate_cost_from_usage,
    sanitize_request_for_logging,
)

logger = logging.getLogger(__name__)


async def aggregate_openai_responses_stream(
    stream: AsyncIterator,
    on_chunk: Callable[[StreamEvent], Awaitable[None]],
) -> tuple[Message, dict[str, Any]]:
    """Aggregate OpenAI Responses API streaming chunks into a complete Message.

    This handles the o1/o3 reasoning models which use a different API than chat completions.
    Key differences:
    - Uses response.output_item.added/done events instead of choices[]
    - Reasoning content comes through response.reasoning_summary_text.delta
    - Different event structure for tool calls and text

    Returns:
        tuple of (final_message, usage_dict)
    """
    assert stream is not None
    assert on_chunk is not None
    assert callable(on_chunk)

    # Emit start event
    await on_chunk(StreamStart())

    # Track content blocks
    content_blocks: list[dict[str, Any]] = []
    current_item: dict[str, Any] | None = None
    current_block_index = -1
    finish_reason = "stop"

    # Usage tracking
    usage_data: dict[str, Any] = {}

    async for event in stream:
        event_type = getattr(event, "type", None)
        if event_type and event_type.startswith("response.reasoning"):
            logger.debug(f"Reasoning event: {event_type}")

        # Handle output item start
        if event_type == "response.output_item.added":
            item = event.item
            item_type = getattr(item, "type", None)

            if item_type == "reasoning":
                # Start thinking block
                current_item = {"type": "reasoning", "summary": []}
                current_block_index = len(content_blocks)
                content_blocks.append({"type": "thinking", "thinking": ""})
                await on_chunk(ThinkingStart(content_index=current_block_index))

            elif item_type == "message":
                # Start text block
                current_item = {"type": "message", "content": []}
                current_block_index = len(content_blocks)
                content_blocks.append({"type": "text", "text": ""})
                await on_chunk(TextStart(content_index=current_block_index))

            elif item_type == "function_call":
                # Start tool call block
                call_id = getattr(item, "call_id", "") + "|" + getattr(item, "id", "")
                name = getattr(item, "name", "")
                current_item = {
                    "type": "function_call",
                    "call_id": call_id,
                    "name": name,
                    "arguments": "",
                }
                current_block_index = len(content_blocks)
                content_blocks.append({
                    "type": "toolCall",
                    "id": call_id,
                    "name": name,
                    "arguments": "",
                })
                await on_chunk(
                    ToolCallStart(
                        content_index=current_block_index,
                        tool_call_id=call_id,
                        tool_name=name,
                    )
                )

        # Handle reasoning summary deltas
        elif event_type == "response.reasoning_summary_part.added":
            if current_item and current_item.get("type") == "reasoning":
                # Push the full part object which includes 'type' field required for re-submission
                # The part has structure like {"type": "summary_text", "text": "..."}
                part_dict = {"type": getattr(event.part, "type", "summary_text"), "text": ""}
                current_item.setdefault("summary", []).append(part_dict)
                logger.debug(f"Added new summary part, total parts: {len(current_item['summary'])}")

        elif event_type == "response.reasoning_summary_text.delta":
            if (
                current_item
                and current_item.get("type") == "reasoning"
                and current_block_index >= 0
            ):
                delta = event.delta
                current_item.setdefault("summary", [])
                if current_item["summary"]:
                    current_item["summary"][-1]["text"] += delta
                content_blocks[current_block_index]["thinking"] += delta
                await on_chunk(
                    ThinkingDelta(
                        content_index=current_block_index,
                        delta=delta,
                    )
                )

        elif event_type == "response.reasoning_summary_part.done":
            # Add newlines between summary parts
            if (
                current_item
                and current_item.get("type") == "reasoning"
                and current_block_index >= 0
            ):
                content_blocks[current_block_index]["thinking"] += "\n\n"
                await on_chunk(
                    ThinkingDelta(
                        content_index=current_block_index,
                        delta="\n\n",
                    )
                )

        # Handle text output deltas
        elif event_type == "response.content_part.added":
            if current_item and current_item.get("type") == "message":
                current_item.setdefault("content", []).append(event.part)

        elif event_type == "response.output_text.delta":
            if current_item and current_item.get("type") == "message" and current_block_index >= 0:
                delta = event.delta
                content_blocks[current_block_index]["text"] += delta
                await on_chunk(
                    TextDelta(
                        content_index=current_block_index,
                        delta=delta,
                    )
                )

        elif event_type == "response.refusal.delta":
            if current_item and current_item.get("type") == "message" and current_block_index >= 0:
                delta = event.delta
                content_blocks[current_block_index]["text"] += delta
                await on_chunk(
                    TextDelta(
                        content_index=current_block_index,
                        delta=delta,
                    )
                )

        # Handle function call argument deltas
        elif event_type == "response.function_call_arguments.delta":
            if (
                current_item
                and current_item.get("type") == "function_call"
                and current_block_index >= 0
            ):
                delta = event.delta
                current_item["arguments"] += delta
                content_blocks[current_block_index]["arguments"] += delta
                partial_args = parse_streaming_json(
                    content_blocks[current_block_index]["arguments"]
                )
                await on_chunk(
                    ToolCallDelta(
                        content_index=current_block_index,
                        tool_call_id=content_blocks[current_block_index]["id"],
                        delta=delta,
                        partial_args=partial_args,
                    )
                )

        # Handle output item completion
        elif event_type == "response.output_item.done":
            item = event.item
            item_type = getattr(item, "type", None)

            if item_type == "reasoning" and current_block_index >= 0:
                # Finalize thinking block - store the full item as signature for re-submission
                thinking_content = content_blocks[current_block_index]["thinking"]
                # Store the raw item as JSON so we can re-submit it exactly to the API
                import json as json_module

                # Build the reasoning item dict using the data we accumulated during streaming
                # The item from the SDK doesn't have the summary we built, so use current_item
                # Only include non-null fields as API rejects null fields (except summary which is required)
                summary = current_item.get("summary", []) if current_item else []
                item_dict = {
                    "type": "reasoning",
                    "id": getattr(item, "id", ""),
                    "summary": summary,  # Required field, even if empty
                }

                content = getattr(item, "content", None)
                if content is not None:
                    item_dict["content"] = content

                encrypted_content = getattr(item, "encrypted_content", None)
                if encrypted_content is not None:
                    item_dict["encrypted_content"] = encrypted_content

                status = getattr(item, "status", None)
                if status is not None:
                    item_dict["status"] = status

                logger.debug(
                    f"Storing reasoning item with {len(summary)} summary parts, fields: {list(item_dict.keys())}"
                )

                # Store the serialized item
                content_blocks[current_block_index]["thinkingSignature"] = json_module.dumps(
                    item_dict
                )

                await on_chunk(
                    ThinkingEnd(
                        content_index=current_block_index,
                        content=thinking_content,
                    )
                )
                current_item = None

            elif item_type == "message" and current_block_index >= 0:
                # Finalize text block - save message ID for re-submission
                text_content = content_blocks[current_block_index]["text"]
                content_blocks[current_block_index]["textSignature"] = getattr(item, "id", None)
                await on_chunk(
                    TextEnd(
                        content_index=current_block_index,
                        content=text_content,
                    )
                )
                current_item = None

            elif item_type == "function_call" and current_block_index >= 0:
                # Finalize tool call block
                try:
                    args = json.loads(getattr(item, "arguments", "{}"))
                    tool_call = ToolCall(
                        id=content_blocks[current_block_index]["id"],
                        name=content_blocks[current_block_index]["name"],
                        args=args,
                    )
                    await on_chunk(
                        ToolCallEnd(
                            content_index=current_block_index,
                            tool_call=tool_call,
                        )
                    )
                except json.JSONDecodeError as e:
                    await on_chunk(
                        ToolCallError(
                            content_index=current_block_index,
                            tool_call_id=content_blocks[current_block_index]["id"],
                            tool_name=content_blocks[current_block_index]["name"],
                            error=f"Invalid JSON arguments: {str(e)}",
                            raw_arguments=content_blocks[current_block_index]["arguments"],
                        )
                    )
                current_item = None

        # Handle completion
        elif event_type == "response.completed":
            response = event.response
            if hasattr(response, "usage") and response.usage:
                # Extract cache tokens from input_tokens_details
                cached_tokens = 0
                if hasattr(response.usage, "input_tokens_details"):
                    details = response.usage.input_tokens_details
                    if hasattr(details, "cached_tokens"):
                        cached_tokens = details.cached_tokens or 0

                # Extract reasoning tokens from output_tokens_details
                reasoning_tokens = 0
                if hasattr(response.usage, "output_tokens_details"):
                    details = response.usage.output_tokens_details
                    if hasattr(details, "reasoning_tokens"):
                        reasoning_tokens = details.reasoning_tokens or 0

                output_tokens = getattr(response.usage, "output_tokens", 0) or 0
                usage_data = {
                    "input_tokens": (getattr(response.usage, "input_tokens", 0) or 0)
                    - cached_tokens,
                    "output_tokens": output_tokens - reasoning_tokens,
                    "reasoning_tokens": reasoning_tokens,
                    "cache_read_tokens": cached_tokens,
                    "cache_write_tokens": 0,
                }

            # Map status to finish reason
            status = getattr(response, "status", "completed")
            if status == "completed":
                finish_reason = "stop"
            elif status == "incomplete":
                finish_reason = "length"
            else:  # failed, cancelled
                finish_reason = "stop"  # Default to stop

            # Override if we have tool calls
            has_tool_calls = any(b.get("type") == "toolCall" for b in content_blocks)
            if has_tool_calls:
                finish_reason = "tool_calls"

        # Handle errors
        elif event_type == "error":
            error_msg = getattr(event, "message", "Unknown error")
            await on_chunk(StreamError(error=error_msg))
            raise Exception(error_msg)

        elif event_type == "response.failed":
            await on_chunk(StreamError(error="Response failed"))
            raise Exception("Response failed")

    # Emit done event
    await on_chunk(StreamDone(finish_reason=finish_reason))

    # Build final message
    # Build final message with ContentBlocks

    final_content_blocks: list = []

    for block in content_blocks:
        if block["type"] == "text":
            final_content_blocks.append(
                TextContent(
                    text=block["text"],
                    text_signature=block.get("textSignature"),  # Message ID for re-submission
                )
            )
        elif block["type"] == "thinking":
            thinking_sig = block.get("thinkingSignature")
            final_content_blocks.append(
                ThinkingContent(
                    thinking=block["thinking"],
                    thinking_signature=thinking_sig,
                )
            )
        elif block["type"] == "toolCall":
            try:
                args = json.loads(block["arguments"]) if block["arguments"] else {}
                if not isinstance(args, dict):
                    raise ValueError(f"Tool args must be object, got {type(args).__name__}")
                final_content_blocks.append(
                    ToolCallContent(
                        id=block["id"],
                        name=block["name"],
                        arguments=args,
                    )
                )
            except (json.JSONDecodeError, ValueError) as e:
                # Include failed tool call with parse error so agent loop can return error to model
                final_content_blocks.append(
                    ToolCallContent(
                        id=block["id"],
                        name=block["name"],
                        arguments={},
                        parse_error=f"Invalid tool arguments: {str(e)}",
                        raw_arguments=block["arguments"],
                    )
                )

    final_message = Message(
        role="assistant",
        content=final_content_blocks,
    )

    logger.debug(f"Built final message with {len(final_content_blocks)} content blocks")

    return final_message, usage_data


def _messages_to_openai_responses(messages: list[Message]) -> list[dict[str, Any]]:
    """Convert rollouts Messages to OpenAI Responses API format.

    Handles new ContentBlock-based message structure:
    - Extracts text from TextContent blocks
    - Extracts thinking from ThinkingContent blocks (with signature for re-use)
    - Converts ToolCallContent to function_call format

    Key differences from chat completions:
    - User messages use input_text content type
    - Assistant messages become separate message/function_call objects
    - Tool calls are separate function_call objects, not properties on messages
    - Tool results become function_call_output objects
    """

    result: list[dict[str, Any]] = []

    for msg in messages:
        if msg.role == "user":
            # Tiger Style: Explicit control flow - handle each content type
            # Handle string content (simple text messages)
            if isinstance(msg.content, str):
                user_text = msg.content
            # Handle ContentBlock list (structured messages)
            elif isinstance(msg.content, list):
                text_blocks = [b for b in msg.content if isinstance(b, TextContent)]
                user_text = "\n".join(b.text for b in text_blocks) if text_blocks else ""
            else:
                raise AssertionError(
                    f"User message content must be str or list[ContentBlock], got {type(msg.content)}"
                )

            result.append({"role": "user", "content": [{"type": "input_text", "text": user_text}]})

        elif msg.role == "assistant":
            # Assistant messages become separate objects
            output: list[dict[str, Any]] = []

            # Tiger Style: Explicit control flow - handle each content type
            # Handle string content (simple text response)
            if isinstance(msg.content, str):
                output.append({
                    "type": "message",
                    "role": "assistant",
                    "content": [{"type": "output_text", "text": msg.content, "annotations": []}],
                    "status": "completed",
                    "id": f"msg_{int(time.time())}_{id(msg)}",
                })
            # Handle ContentBlock list (structured response with thinking/tools)
            elif isinstance(msg.content, list):
                # Process ContentBlocks
                for block in msg.content:
                    if isinstance(block, ThinkingContent) and block.thinking_signature:
                        # Reuse existing reasoning item only if it has encrypted_content
                        # Without encrypted_content, the API rejects the reasoning item
                        # See: https://community.openai.com/t/need-reasoning-false-option-for-gpt-5/1351588/7
                        try:
                            reasoning_item = json.loads(block.thinking_signature)
                            if reasoning_item.get("encrypted_content"):
                                output.append(reasoning_item)
                                logger.debug(
                                    f"Added reasoning item with encrypted_content: id={reasoning_item.get('id', 'unknown')}"
                                )
                            else:
                                logger.debug(
                                    f"Skipping reasoning item without encrypted_content: id={reasoning_item.get('id', 'unknown')}"
                                )
                        except json.JSONDecodeError as e:
                            logger.exception(f"Failed to parse thinking_signature: {e}")
                            pass
                    elif isinstance(block, TextContent):
                        # Add text content as message object
                        # Use actual message ID from API if available (for proper re-submission)
                        msg_id = block.text_signature or f"msg_{int(time.time())}_{id(msg)}"
                        output.append({
                            "type": "message",
                            "role": "assistant",
                            "content": [
                                {"type": "output_text", "text": block.text, "annotations": []}
                            ],
                            "status": "completed",
                            "id": msg_id,
                        })
                    elif isinstance(block, ToolCallContent):
                        # Tool call IDs in responses API use format: call_id|id
                        # Split if already in that format, otherwise use as call_id
                        if "|" in block.id:
                            call_id, func_id = block.id.split("|", 1)
                        else:
                            call_id = block.id
                            func_id = f"fc_{int(time.time())}"

                        output.append({
                            "type": "function_call",
                            "id": func_id,
                            "call_id": call_id,
                            "name": block.name,
                            "arguments": json.dumps(block.arguments),
                        })

            # Add all output objects
            if output:
                result.extend(output)

        elif msg.role == "tool":
            # Extract text from ContentBlocks for tool result
            # Tiger Style: Explicit control flow - handle both string and ContentBlock content
            if isinstance(msg.content, str):
                tool_result_text = msg.content
            elif isinstance(msg.content, list):
                text_blocks = [b for b in msg.content if isinstance(b, TextContent)]
                tool_result_text = "\n".join(b.text for b in text_blocks) if text_blocks else ""
            else:
                tool_result_text = ""

            # Extract call_id from tool_call_id (format: call_id|id or just call_id)
            call_id = msg.tool_call_id
            if "|" in call_id:
                call_id = call_id.split("|")[0]

            result.append({
                "type": "function_call_output",
                "call_id": call_id,
                "output": tool_result_text,
            })

    return result


async def rollout_openai_responses(
    actor: Actor,
    on_chunk: Callable[[StreamEvent], Awaitable[None]],
    **kwargs: Any,
) -> Actor:
    """Make an OpenAI Responses API call (for o1/o3 reasoning models) with streaming.

    This uses the Responses API which is different from the Chat Completions API.
    It's specifically designed for reasoning models that produce extended thinking.

    Note: **kwargs accepts but ignores provider-specific params (e.g., anthropic thinking params)
    """
    assert actor is not None
    assert isinstance(actor, Actor)
    assert actor.endpoint is not None
    assert actor.trajectory is not None
    assert on_chunk is not None
    assert callable(on_chunk)

    client_kwargs = {
        "api_key": actor.endpoint.api_key,
        "max_retries": actor.endpoint.max_retries,
        "timeout": actor.endpoint.timeout,
    }
    # Only set base_url if it's provided (non-empty)
    if actor.endpoint.api_base:
        client_kwargs["base_url"] = actor.endpoint.api_base

    # Transform messages for cross-provider compatibility (like pi-ai does)
    from ..transform_messages import transform_messages

    transformed_messages = transform_messages(
        actor.trajectory.messages,
        target_provider=actor.endpoint.provider,
        target_api="openai-responses",
    )

    # Strip details before sending to LLM
    llm_messages = _prepare_messages_for_llm(transformed_messages)

    # Convert messages to OpenAI Responses format
    # Note: The Responses API uses a completely different message format than chat completions
    messages = _messages_to_openai_responses(llm_messages)

    params = {
        "model": actor.endpoint.model,
        "input": messages,  # Note: Responses API uses 'input' not 'messages'
        "stream": True,
    }

    # Add max_output_tokens (not max_completion_tokens for Responses API)
    if actor.endpoint.max_completion_tokens is not None:
        params["max_output_tokens"] = actor.endpoint.max_completion_tokens
    elif actor.endpoint.max_tokens is not None:
        params["max_output_tokens"] = actor.endpoint.max_tokens

    # Temperature is supported in Responses API (but not for some reasoning models like GPT-5-Codex)
    # Skip temperature for GPT-5 models which don't support it
    model_name = actor.endpoint.model.lower()
    if hasattr(actor.endpoint, "temperature") and actor.endpoint.temperature is not None:
        if not model_name.startswith("gpt-5"):
            params["temperature"] = actor.endpoint.temperature

    # Add reasoning config for reasoning models
    # Check if model supports reasoning
    from ..models import get_model

    try:
        model_metadata = get_model(actor.endpoint.provider, actor.endpoint.model)
        is_reasoning_model = model_metadata and model_metadata.reasoning
    except (KeyError, ValueError):
        is_reasoning_model = False

    # GPT-5 models always produce reasoning, so we must always request encrypted_content
    # to be able to re-submit the reasoning items in subsequent turns
    # See: https://community.openai.com/t/need-reasoning-false-option-for-gpt-5/1351588/7
    if is_reasoning_model and model_name.startswith("gpt-5"):
        if actor.endpoint.reasoning_effort is not None:
            params["reasoning"] = {
                "effort": actor.endpoint.reasoning_effort,
                "summary": "auto",
            }
        else:
            # Default to low effort to minimize cost when not explicitly requested
            params["reasoning"] = {
                "effort": "low",
                "summary": "auto",
            }
        params["include"] = ["reasoning.encrypted_content"]
    elif actor.endpoint.reasoning_effort is not None and is_reasoning_model:
        # Non-GPT-5 reasoning models with explicit reasoning_effort
        params["reasoning"] = {
            "effort": actor.endpoint.reasoning_effort,
            "summary": "auto",
        }
        params["include"] = ["reasoning.encrypted_content"]

    # Tools for Responses API (different format than chat completions!)
    if actor.tools:
        # Responses API expects flat structure: {type, name, description, parameters}
        # Not nested like chat completions: {type, function: {name, description, parameters}}
        params["tools"] = [
            {
                "type": "function",
                "name": t.function.name,
                "description": t.function.description,
                "parameters": {
                    "type": t.function.parameters.type,
                    "properties": t.function.parameters.properties,
                    "required": t.function.required,
                },
                "strict": None,
            }
            for t in actor.tools
        ]

    if hasattr(actor.endpoint, "extra_params") and actor.endpoint.extra_params:
        params.update(actor.endpoint.extra_params)

    client = AsyncOpenAI(**client_kwargs)

    # Ensure client is closed even on exception. This prevents httpx connection cleanup
    # from racing with trio_asyncio teardown, which causes "Task got bad yield" errors.
    # Alternative fix: scope trio_asyncio.open_loop() to only SSH/asyncio operations so httpx
    # runs in pure trio, but that's a larger refactor - this try/finally is sufficient.
    try:
        # Use the responses.create endpoint
        stream = await client.responses.create(**params)
        final_message, usage_data = await aggregate_openai_responses_stream(stream, on_chunk)

    except Exception as e:
        from openai import BadRequestError, RateLimitError

        from ..store import log_crash

        sanitized = sanitize_request_for_logging(params)

        if isinstance(e, BadRequestError):
            crash_file = log_crash(e, "openai_responses", actor.endpoint.model)
            raise AssertionError(
                f"API returned 400 Bad Request: {e}\nCrash details written to: {crash_file}"
            ) from e

        if isinstance(e, RateLimitError):
            from .base import ProviderError

            logger.warning(f"Rate limit exceeded for {actor.endpoint.model}")
            raise ProviderError(
                f"OpenAI Responses rate limit exceeded: {e}",
                original_error=e,
                attempts=actor.endpoint.max_retries,
                provider="openai",
            ) from e

        # For other transient errors, wrap as ProviderError
        import httpx
        from openai import APIConnectionError, APITimeoutError, InternalServerError

        # Transient network/API errors that should trigger retry
        if isinstance(
            e, (APIConnectionError, APITimeoutError, InternalServerError, httpx.RemoteProtocolError)
        ):
            from .base import ProviderError

            logger.exception(
                f"OpenAI Responses API call failed: {e}\n  Model: {actor.endpoint.model}",
                extra={
                    "exception": str(e),
                    "request_params": sanitized,
                    "model": actor.endpoint.model,
                },
            )
            raise ProviderError(
                f"OpenAI Responses API error: {e}",
                original_error=e,
                attempts=actor.endpoint.max_retries,
                provider="openai",
            ) from e

        # Other errors - log and re-raise as-is
        logger.exception(
            f"OpenAI Responses API call failed: {e}\n  Model: {actor.endpoint.model}",
            extra={
                "exception": str(e),
                "request_params": sanitized,
                "model": actor.endpoint.model,
            },
        )
        raise
    finally:
        await client.close()

    # Build completion object with granular token breakdown
    usage = Usage(
        input_tokens=usage_data.get("input_tokens", 0),
        output_tokens=usage_data.get("output_tokens", 0),
        reasoning_tokens=usage_data.get("reasoning_tokens", 0),
        cache_read_tokens=usage_data.get("cache_read_tokens", 0),
        cache_write_tokens=usage_data.get("cache_write_tokens", 0),
    )

    # Enrich message with provider/api/model metadata for cross-provider handoff
    final_message = replace(
        final_message,
        provider=actor.endpoint.provider,
        api="openai-responses",
        model=actor.endpoint.model,
    )

    # Calculate cost if model pricing is available
    from ..models import get_model

    model_meta = get_model(actor.endpoint.provider, actor.endpoint.model)
    if model_meta and model_meta.cost:
        cost = calculate_cost_from_usage(usage, model_meta.cost)
        usage = replace(usage, cost=cost)

    completion = ChatCompletion(
        id="responses-" + str(int(time.time())),
        object="chat.completion",
        created=int(time.time()),
        model=actor.endpoint.model,
        usage=usage,
        choices=[Choice(0, final_message, "stop")],
    )

    new_trajectory = replace(
        actor.trajectory,
        messages=actor.trajectory.messages + [final_message],
        completions=actor.trajectory.completions + [completion],
    )

    result_actor = replace(actor, trajectory=new_trajectory)
    assert result_actor is not None
    assert result_actor.trajectory is not None
    return result_actor



"""Utilities for aggregating trials by configuration."""

import math
from collections import defaultdict

from wafer_core.tools.autotuner.dtypes import AggregatedConfig, MetricStats, Trial, TrialStatus


def aggregate_trials_by_config(
    trials: list[Trial], trials_per_config: int
) -> list[AggregatedConfig]:
    """Group trials by configuration and compute aggregated statistics.

    Args:
        trials: List of all trials
        trials_per_config: Number of trials per configuration

    Returns:
        List of aggregated configurations with statistics
    """
    if trials_per_config == 1:
        # No aggregation needed - return empty list
        # Caller should use trials directly
        return []

    # Group trials by config_number
    config_groups: dict[int, list[Trial]] = defaultdict(list)
    for trial in trials:
        config_number = trial.trial_number // trials_per_config
        config_groups[config_number].append(trial)

    # Compute aggregated statistics for each config
    aggregated_configs: list[AggregatedConfig] = []

    for config_number, config_trials in sorted(config_groups.items()):
        # Only include successful trials in aggregation
        successful_trials = [
            t for t in config_trials if t.status == TrialStatus.SUCCESS
        ]

        if not successful_trials:
            # Skip configs with no successful trials
            continue

        # Use config from first trial (should be same for all trials in group)
        config = successful_trials[0].config

        # Collect all metric names
        all_metrics = set()
        for trial in successful_trials:
            all_metrics.update(trial.metrics.keys())

        # Aggregate metrics
        aggregated_metrics: dict[str, MetricStats] = {}

        for metric in all_metrics:
            # Get all values for this metric
            values = [
                trial.metrics[metric]
                for trial in successful_trials
                if metric in trial.metrics and isinstance(trial.metrics[metric], (int, float))
            ]

            if values:
                # Compute statistics
                mean = sum(values) / len(values)
                variance = sum((v - mean) ** 2 for v in values) / len(values)
                std = math.sqrt(variance)
                min_val = min(values)
                max_val = max(values)

                aggregated_metrics[metric] = MetricStats(
                    mean=mean, std=std, min=min_val, max=max_val, values=values
                )

        # Check if all trials passed constraints
        all_passed = all(t.passed_constraints for t in successful_trials)

        # Check if any trial is pareto optimal
        is_pareto = any(t.is_pareto_optimal for t in successful_trials)

        aggregated_configs.append(
            AggregatedConfig(
                config_number=config_number,
                config=config,
                trials=config_trials,
                metrics=aggregated_metrics,
                all_passed_constraints=all_passed,
                is_pareto_optimal=is_pareto,
            )
        )

    return sorted(aggregated_configs, key=lambda c: c.config_number)



"""Trace loading and parsing logic.

Loads JSON trace files from AMD/NVIDIA profilers and extracts kernel execution data,
Python call stacks, CPU operator mappings, and layer correlations.
"""

import bisect
import sys
from collections import defaultdict
from collections.abc import Callable
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any

try:
    import ijson
except ImportError:
    ijson = None

import orjson
import pandas as pd

from .architecture import detect_architecture
from .classifier import classify
from .layer_segmentation import segment_layers_by_architecture

ProgressCallback = Callable[[str, float], None]


@dataclass
class SinglePassResult:
    """Collected data from single-pass event processing."""
    cpu_op_mapping: dict[int, str] = field(default_factory=dict)
    python_intervals: list[tuple[int, int, int, int | None, str]] = field(default_factory=list)
    # Raw python events for lazy python_by_id construction (built on-demand in _get_python_stack_full)
    python_events_raw: list[dict[str, Any]] = field(default_factory=list)
    phases: list[dict[str, Any]] = field(default_factory=list)
    correlation_groups: dict[int, dict[str, Any]] = field(default_factory=lambda: defaultdict(
        lambda: {"count": 0, "has_attention": False, "has_ffn": False}
    ))
    kernel_events: list[dict[str, Any]] = field(default_factory=list)
    # Lazily built when needed for stack resolution
    _python_by_id: dict[int, dict[str, Any]] | None = field(default=None)
    
    @property
    def python_by_id(self) -> dict[int, dict[str, Any]]:
        """Lazily build python_by_id from raw events on first access."""
        if self._python_by_id is None:
            self._python_by_id = {}
            for ev in self.python_events_raw:
                args = ev.get("args")
                py_id = args.get("Python id") if args else None
                if py_id is not None:
                    self._python_by_id[py_id] = {
                        "name": ev["name"],
                        "parent_id": args.get("Python parent id") if args else None,
                    }
        return self._python_by_id


@dataclass
class LoadedTrace:
    """Complete trace data loaded once and reused by all analysis functions."""
    platform: str
    gpu_name: str
    device_props: dict[str, Any]
    df: pd.DataFrame
    patterns: dict[tuple[str, str], set[str]]
    layers: dict[int, int]
    # For fusion/warnings (kept from raw JSON)
    kernel_events: list[dict[str, Any]]
    all_events: list[dict[str, Any]]
    correlation_groups: dict[int, list[dict[str, Any]]]
    phases: list[dict[str, Any]]  # Phase annotations for alignment


def _process_events_single_pass(
    events: list[dict[str, Any]],
    include_stacks: bool = True,
) -> SinglePassResult:
    """Process all events in a single iteration.
    
    Optimizations applied:
    - Cache list.append methods for 2-3x speedup on hot paths
    - Store raw python events, build python_by_id lazily (only ~48 lookups due to caching)
    - Local variable caching for frequently accessed attributes
    
    Args:
        events: List of trace events
        include_stacks: Whether to collect Python stack info (expensive operation).
                       When False, skips python_function processing entirely.
    """
    result = SinglePassResult()
    correlation_groups: dict[int, dict[str, Any]] = defaultdict(
        lambda: {"count": 0, "has_attention": False, "has_ffn": False}
    )
    
    # Cache list.append methods for faster appending (measured 2-3x speedup)
    kernel_append = result.kernel_events.append
    python_interval_append = result.python_intervals.append
    python_raw_append = result.python_events_raw.append
    phase_append = result.phases.append
    cpu_op_mapping = result.cpu_op_mapping
    
    for ev in events:
        cat = ev.get("cat")
        
        if cat == "kernel":
            args = ev.get("args")
            corr_id = args.get("correlation") if args else None
            if corr_id is not None:
                kernel_name = ev.get("name", "").lower()
                grp = correlation_groups[corr_id]
                grp["count"] += 1
                if "attention" in kernel_name or "fmha" in kernel_name:
                    grp["has_attention"] = True
                if "cijk_" in kernel_name or "nvjet" in kernel_name or "wvsplitk" in kernel_name or "gemm" in kernel_name:
                    grp["has_ffn"] = True
            kernel_append(ev)
        
        elif cat == "cpu_op":
            args = ev.get("args")
            ext_id = args.get("External id") if args else None
            if ext_id is not None:
                cpu_op_mapping[ext_id] = ev.get("name", "")
        
        elif cat == "python_function" and include_stacks:
            # Store raw event for lazy python_by_id construction
            python_raw_append(ev)
            # Build interval tuple for binary search
            args = ev.get("args")
            py_id = args.get("Python id") if args else None
            ts = ev["ts"]
            dur = ev.get("dur", 0)
            python_interval_append((ts, ts + dur, dur, py_id, ev["name"]))
                
        elif cat == "user_annotation":
            name = ev.get("name", "")
            if name.startswith("execute_context"):
                tokens = 0
                parts = name.split("_")
                for i, p in enumerate(parts):
                    if i > 0 and parts[i-1] == "context" and "(" in p and ")" in p:
                        try:
                            tokens = int(p.split("(")[1].split(")")[0])
                            break
                        except Exception:
                            pass
                is_prefill = "generation_0" in name and tokens > 0
                phase_append({
                    "type": "prefill" if is_prefill else "decode",
                    "ts_start": ev["ts"],
                    "ts_end": ev["ts"] + ev["dur"],
                })
    
    if include_stacks and result.python_intervals:
        result.python_intervals.sort()
    
    result.correlation_groups = dict(correlation_groups)
    
    return result


def _build_layer_mapping(correlation_groups: dict[int, dict[str, Any]]) -> dict[int, int]:
    """Build layer mapping from correlation groups."""
    correlation_to_layer = {}
    layer_num = 0
    
    for corr_id in sorted(correlation_groups.keys()):
        group = correlation_groups[corr_id]
        
        is_layer = (
            group["count"] >= 70 and group["has_attention"] and group["has_ffn"]
        )
        
        if is_layer:
            correlation_to_layer[corr_id] = layer_num
            layer_num += 1
    
    return correlation_to_layer


def _get_python_stack_full(
    timestamp: int,
    python_intervals: list[tuple[int, int, int, int | None, str]],
    python_by_id: dict[int, dict[str, Any]],
) -> tuple[str | None, list[str]]:
    """Get full Python call stack for a kernel launch.
    
    Computes the chain on-demand by walking parent pointers.
    This is fast because we only call this ~48 times due to cpu_op caching.
    """
    idx = bisect.bisect_right(
        python_intervals, (timestamp, float("inf"), float("inf"), None, "")
    )
    
    active_funcs = []
    for i in range(idx - 1, max(0, idx - 1000), -1):
        ts_start, ts_end, duration, py_id, name = python_intervals[i]
        if ts_start <= timestamp <= ts_end:
            active_funcs.append((duration, py_id, name))
        if ts_end < timestamp - 1000000:
            break
    
    if not active_funcs:
        return None, []
    
    active_funcs.sort()
    leaf_duration, leaf_id, leaf_name = active_funcs[0]
    
    full_stack = []
    current_id = leaf_id
    visited: set[int] = set()
    
    while current_id is not None and current_id not in visited and current_id in python_by_id:
        func = python_by_id[current_id]
        full_stack.append(func["name"])
        visited.add(current_id)
        current_id = func["parent_id"]
        if len(full_stack) >= 50:
            break
    
    full_stack.reverse()
    
    summary = None
    vllm_funcs = [f for f in full_stack if any(x in f.lower() for x in ["vllm/", "model", "<eval_with_key>"])]
    
    if vllm_funcs:
        summary = vllm_funcs[-1]
        if any("torch/cuda/graphs" in f for f in full_stack):
            if len(summary) > 45:
                summary = "vllm/..." + summary.split("/")[-1]
            summary = f"{summary} [CUDA graph]"
        elif len(summary) > 53:
            summary = "vllm/..." + summary.split("/")[-1]
    else:
        summary = leaf_name
    
    return summary, full_stack


def load_trace(
    file_path: str | Path,
    include_stacks: bool = True,
    on_progress: ProgressCallback | None = None,
) -> tuple[str, str, dict[str, Any], pd.DataFrame, dict[tuple[str, str], set[str]], dict[int, int]]:
    """Load trace and return platform info, device properties, kernels, patterns, and layer mapping.
    
    Args:
        file_path: Path to the trace JSON file
        include_stacks: Whether to resolve Python call stacks (slower but more info)
        on_progress: Optional callback for progress updates: (stage_name, progress_fraction)
    """
    def _progress(stage: str, pct: float) -> None:
        if on_progress:
            on_progress(stage, pct)
    
    _progress("Reading file", 0.0)
    with open(file_path, "rb") as f:
        raw = f.read()
    
    _progress("Parsing JSON", 0.1)
    trace = orjson.loads(raw)
    
    props = trace.get("deviceProperties", [{}])[0]
    is_amd = trace.get("roctracer_version") or props.get("warpSize") == 64
    platform = "AMD" if is_amd else "NVIDIA"
    gpu_name = props.get("name", "MI300X" if is_amd else "Unknown GPU")
    
    device_props = {
        "name": gpu_name,
        "compute_capability": f"{props.get('computeMajor', 0)}.{props.get('computeMinor', 0)}",
        "total_memory_gb": props.get("totalGlobalMem", 0) / (1024**3),
        "sm_count": props.get("numSms", 0),
        "warp_size": props.get("warpSize", 32),
        "max_threads_per_block": props.get("maxThreadsPerBlock", 0),
        "shared_mem_per_block_kb": props.get("sharedMemPerBlock", 0) / 1024,
    }
    
    _progress("Processing events", 0.4)
    events = trace.get("traceEvents", [])
    pass_result = _process_events_single_pass(events, include_stacks=include_stacks)
    
    _progress("Detecting architecture", 0.6)
    kernel_names = [ev.get("name", "") for ev in pass_result.kernel_events]
    architecture, _ = detect_architecture(kernel_names)
    
    layer_kernels_dict, layer_warnings = segment_layers_by_architecture(
        pass_result.kernel_events,
        architecture,
    )
    
    # Convert layer_kernels_dict (layer_num -> kernels) to correlation_id -> layer_num mapping
    # Fall back to old method if architecture-based segmentation fails
    if layer_kernels_dict:
        layer_mapping: dict[int, int] = {}
        for layer_num, kernels in layer_kernels_dict.items():
            for kernel in kernels:
                corr_id = kernel.get("args", {}).get("correlation")
                if corr_id is not None:
                    layer_mapping[corr_id] = layer_num
    else:
        # Fallback to correlation-based method if architecture segmentation failed
        layer_mapping = _build_layer_mapping(pass_result.correlation_groups)
    
    kernel_data = []
    kernel_patterns: dict[tuple[str, str], set[str]] = defaultdict(set)
    sorted_phases = sorted(pass_result.phases, key=lambda p: p["ts_start"])
    
    phase_starts = [p["ts_start"] for p in sorted_phases]
    phase_types = [p["type"] for p in sorted_phases]
    phase_ends = [p["ts_end"] for p in sorted_phases]
    
    def _get_phase_for_timestamp(ts: int) -> str:
        """Get phase for a timestamp using binary search. O(log n)."""
        if not phase_starts:
            return "decode"
        idx = bisect.bisect_right(phase_starts, ts) - 1
        if idx >= 0 and phase_starts[idx] <= ts <= phase_ends[idx]:
            return phase_types[idx]
        return "decode"
    
    cpu_op_cache: dict[str, str | None] = {}
    
    _progress("Classifying kernels", 0.7)
    for ev in pass_result.kernel_events:
        name_raw = ev["name"]
        name = sys.intern(name_raw)
        dur, ts = ev.get("dur", 0), ev["ts"]
        corr_id = ev.get("args", {}).get("correlation")
        ext_id = ev.get("args", {}).get("External id")
        
        phase = _get_phase_for_timestamp(ts)
        
        op, pattern = classify(name, platform)
        kernel_patterns[(op.value, phase)].add(pattern)
        
        layer = layer_mapping.get(corr_id) if corr_id is not None else None
        cpu_op = pass_result.cpu_op_mapping.get(ext_id) if ext_id is not None else None
        python_stack: list[str] = []
        
        if cpu_op is None and include_stacks:
            if name in cpu_op_cache:
                cpu_op = cpu_op_cache[name]
            else:
                cpu_op, python_stack = _get_python_stack_full(
                    ts, pass_result.python_intervals, pass_result.python_by_id
                )
                cpu_op_cache[name] = cpu_op
        
        kernel_data.append({
            "name": name,
            "dur_us": dur,
            "phase": phase,
            "op": op.value,
            "pattern": pattern,
            "layer": layer,
            "correlation": corr_id,
            "cpu_op": cpu_op,
            "python_stack": python_stack,
        })
    
    _progress("Building DataFrame", 0.95)
    df = pd.DataFrame(kernel_data)
    _progress("Complete", 1.0)
    
    return platform, gpu_name, device_props, df, dict(kernel_patterns), layer_mapping


def load_trace_full(
    file_path: str | Path,
    include_stacks: bool = True,
    on_progress: ProgressCallback | None = None,
) -> LoadedTrace:
    """Load trace once with all data needed by downstream analysis functions.
    
    Args:
        file_path: Path to the trace JSON file
        include_stacks: Whether to resolve Python call stacks
        on_progress: Optional callback for progress updates
        
    Returns:
        LoadedTrace with all trace data
    """
    def _progress(stage: str, pct: float) -> None:
        if on_progress:
            on_progress(stage, pct)
    
    _progress("Reading file", 0.0)
    with open(file_path, "rb") as f:
        raw = f.read()
    
    _progress("Parsing JSON", 0.1)
    trace = orjson.loads(raw)
    all_events = trace.get("traceEvents", [])
    
    props = trace.get("deviceProperties", [{}])[0]
    is_amd = trace.get("roctracer_version") or props.get("warpSize") == 64
    platform = "AMD" if is_amd else "NVIDIA"
    gpu_name = props.get("name", "MI300X" if is_amd else "Unknown GPU")
    
    device_props = {
        "name": gpu_name,
        "compute_capability": f"{props.get('computeMajor', 0)}.{props.get('computeMinor', 0)}",
        "total_memory_gb": props.get("totalGlobalMem", 0) / (1024**3),
        "sm_count": props.get("numSms", 0),
        "warp_size": props.get("warpSize", 32),
        "max_threads_per_block": props.get("maxThreadsPerBlock", 0),
        "shared_mem_per_block_kb": props.get("sharedMemPerBlock", 0) / 1024,
    }
    
    _progress("Processing events", 0.4)
    pass_result = _process_events_single_pass(all_events, include_stacks=include_stacks)
    
    _progress("Detecting architecture", 0.6)
    kernel_names = [ev.get("name", "") for ev in pass_result.kernel_events]
    architecture, _ = detect_architecture(kernel_names)
    
    layer_kernels_dict, layer_warnings = segment_layers_by_architecture(
        pass_result.kernel_events,
        architecture,
    )
    
    # Convert layer_kernels_dict (layer_num -> kernels) to correlation_id -> layer_num mapping
    if layer_kernels_dict:
        layer_mapping: dict[int, int] = {}
        for layer_num, kernels in layer_kernels_dict.items():
            for kernel in kernels:
                corr_id = kernel.get("args", {}).get("correlation")
                if corr_id is not None:
                    layer_mapping[corr_id] = layer_num
    else:
        layer_mapping = _build_layer_mapping(pass_result.correlation_groups)
    
    kernel_data = []
    kernel_patterns: dict[tuple[str, str], set[str]] = defaultdict(set)
    sorted_phases = sorted(pass_result.phases, key=lambda p: p["ts_start"])
    
    phase_starts = [p["ts_start"] for p in sorted_phases]
    phase_types = [p["type"] for p in sorted_phases]
    phase_ends = [p["ts_end"] for p in sorted_phases]
    
    def _get_phase_for_timestamp(ts: int) -> str:
        """Get phase for a timestamp using binary search. O(log n)."""
        if not phase_starts:
            return "decode"
        idx = bisect.bisect_right(phase_starts, ts) - 1
        if idx >= 0 and phase_starts[idx] <= ts <= phase_ends[idx]:
            return phase_types[idx]
        return "decode"
    
    cpu_op_cache: dict[str, str | None] = {}
    
    _progress("Classifying kernels", 0.7)
    for ev in pass_result.kernel_events:
        name_raw = ev["name"]
        name = sys.intern(name_raw)
        dur, ts = ev.get("dur", 0), ev["ts"]
        corr_id = ev.get("args", {}).get("correlation")
        ext_id = ev.get("args", {}).get("External id")
        
        phase = _get_phase_for_timestamp(ts)
        
        op, pattern = classify(name, platform)
        kernel_patterns[(op.value, phase)].add(pattern)
        
        layer = layer_mapping.get(corr_id) if corr_id is not None else None
        cpu_op = pass_result.cpu_op_mapping.get(ext_id) if ext_id is not None else None
        python_stack: list[str] = []
        
        if cpu_op is None and include_stacks:
            if name in cpu_op_cache:
                cpu_op = cpu_op_cache[name]
            else:
                cpu_op, python_stack = _get_python_stack_full(
                    ts, pass_result.python_intervals, pass_result.python_by_id
                )
                cpu_op_cache[name] = cpu_op
        
        kernel_data.append({
            "name": name,
            "dur_us": dur,
            "phase": phase,
            "op": op.value,
            "pattern": pattern,
            "layer": layer,
            "correlation": corr_id,
            "cpu_op": cpu_op,
            "python_stack": python_stack,
        })
    
    _progress("Building DataFrame", 0.95)
    df = pd.DataFrame(kernel_data)
    
    kernel_events = pass_result.kernel_events
    correlation_groups: dict[int, list[dict[str, Any]]] = defaultdict(list)
    for ev in kernel_events:
        corr_id = ev.get("args", {}).get("correlation")
        if corr_id is not None:
            correlation_groups[corr_id].append(ev)
    
    _progress("Complete", 1.0)
    
    return LoadedTrace(
        platform=platform,
        gpu_name=gpu_name,
        device_props=device_props,
        df=df,
        patterns=dict(kernel_patterns),
        layers=layer_mapping,
        kernel_events=kernel_events,
        all_events=all_events,
        correlation_groups=dict(correlation_groups),
        phases=pass_result.phases,
    )


@dataclass
class StreamingMetadata:
    """Early metadata available before full trace processing."""
    platform: str
    gpu_name: str
    device_props: dict[str, Any]
    file_size_mb: float


def _extract_metadata_fast(file_path: Path) -> StreamingMetadata:
    """Extract trace metadata instantly using streaming parser (~2ms).
    
    Uses ijson to read only the deviceProperties section without
    parsing the entire file. Falls back to full parse if ijson unavailable.
    """
    file_size_mb = file_path.stat().st_size / (1024 * 1024)
    platform = "Unknown"
    gpu_name = "Unknown GPU"
    device_props: dict[str, Any] = {}
    
    if ijson is None:
        # Fallback: parse full file (slower but works)
        with open(file_path, "rb") as f:
            trace = orjson.loads(f.read())
        props = trace.get("deviceProperties", [{}])[0]
        is_amd = trace.get("roctracer_version") or props.get("warpSize") == 64
        platform = "AMD" if is_amd else "NVIDIA"
        gpu_name = props.get("name", "MI300X" if is_amd else "Unknown GPU")
        device_props = {
            "name": gpu_name,
            "compute_capability": f"{props.get('computeMajor', 0)}.{props.get('computeMinor', 0)}",
            "total_memory_gb": props.get("totalGlobalMem", 0) / (1024**3),
            "sm_count": props.get("numSms", 0),
            "warp_size": props.get("warpSize", 32),
            "max_threads_per_block": props.get("maxThreadsPerBlock", 0),
            "shared_mem_per_block_kb": props.get("sharedMemPerBlock", 0) / 1024,
        }
        return StreamingMetadata(
            platform=platform,
            gpu_name=gpu_name,
            device_props=device_props,
            file_size_mb=file_size_mb,
        )
    
    with open(file_path, "rb") as f:
        parser = ijson.parse(f)
        for prefix, event, value in parser:
            if prefix == "deviceProperties.item.name":
                gpu_name = value
            elif prefix == "deviceProperties.item.warpSize":
                platform = "AMD" if value == 64 else "NVIDIA"
            elif prefix == "deviceProperties.item.computeMajor":
                device_props["compute_major"] = value
            elif prefix == "deviceProperties.item.computeMinor":
                device_props["compute_minor"] = value
            elif prefix == "deviceProperties.item.totalGlobalMem":
                device_props["total_memory_gb"] = value / (1024**3)
            elif prefix == "deviceProperties.item.numSms":
                device_props["sm_count"] = value
            elif prefix == "deviceProperties.item.maxThreadsPerBlock":
                device_props["max_threads_per_block"] = value
            elif prefix == "deviceProperties.item.sharedMemPerBlock":
                device_props["shared_mem_per_block_kb"] = value / 1024
            elif prefix == "traceEvents.item":
                # Hit first event, stop - we only need metadata
                break
    
    # Fallback platform detection
    if platform == "Unknown":
        platform = "AMD" if "MI" in gpu_name or "Instinct" in gpu_name else "NVIDIA"
    
    device_props["name"] = gpu_name
    device_props["compute_capability"] = f"{device_props.get('compute_major', 0)}.{device_props.get('compute_minor', 0)}"
    device_props["warp_size"] = 64 if platform == "AMD" else 32
    
    return StreamingMetadata(
        platform=platform,
        gpu_name=gpu_name,
        device_props=device_props,
        file_size_mb=file_size_mb,
    )


def load_trace_streaming(
    file_path: str | Path,
    include_stacks: bool = True,
    on_metadata: Callable[[StreamingMetadata], None] | None = None,
    on_progress: ProgressCallback | None = None,
) -> tuple[str, str, dict[str, Any], pd.DataFrame, dict[tuple[str, str], set[str]], dict[int, int]]:
    """Load trace with instant metadata feedback.
    
    Hybrid approach:
    1. Phase 1 (~2ms): Extract metadata with ijson streaming
    2. Phase 2 (full): Parse with orjson and process
    
    The on_metadata callback fires immediately with GPU/platform info,
    allowing the UI to show feedback while the full load continues.
    
    Args:
        file_path: Path to the trace JSON file
        include_stacks: Whether to resolve Python call stacks
        on_metadata: Callback for instant metadata (fires in ~2ms)
        on_progress: Callback for progress updates during processing
    """
    file_path = Path(file_path)
    
    # Phase 1: Instant metadata extraction (~2ms)
    metadata = _extract_metadata_fast(file_path)
    
    if on_metadata:
        on_metadata(metadata)
    
    # Phase 2: Full load with orjson (fast) + progress updates
    return load_trace(
        file_path,
        include_stacks=include_stacks,
        on_progress=on_progress,
    )



"""Search strategies for generating trial configurations."""

from typing import Any

from wafer_core.tools.autotuner.dtypes import SearchSpace


def generate_grid_trials(search_space: SearchSpace, max_trials: int | None = None) -> list[dict[str, Any]]:
    """Generate trial configs via grid search (Cartesian product).

    Args:
        search_space: Parameter space definition
        max_trials: Optional limit on number of trials

    Returns:
        List of trial configurations (dicts of param -> value)
    """
    configs = search_space.grid_configs()

    if max_trials is not None and len(configs) > max_trials:
        configs = configs[:max_trials]

    return configs
